/***************************************************************************************
* File name: T_S_fm.c
* Purpose:  Eliminates the Temporal and the Spatial Redundancy.
*
* Description:	The Temporal Redundancy Reduction Program is Used to Reduce
*				(Remove) the Correlation of the Data in Processing Buffer. The
*				Spatial Redundancy Reduction Packs Again the Samples in 16 Bits
****************************************************************************************/

#ifdef SIMULATOR
#include "virtuosoSim.h"
#include "stdio.h"
#else // SIMULATOR
#include "node1.h"
#include "k_struct.h"
#endif
#include "spuAsw.h"
#include "spulib.h" 


/*=================================================================================================*/
/*Functions declared in this file
===================================================================================================*/

void Frameify (int *src, int srcSizeInWords, int pixelsPerFrame, int *dest);
void Keyframe (int *src, int srcSizeInWords, int pixelsPerFrame, int *dest);
void _Delta (int *buf, int words);
void Pixelize (int *src, int srcSizeInWords, int pixelsPerFrame, int *dest);
void Map2Pos32 (int *buf, int sizeInWords);
int RampDiff (u4 *buf, int size_w, int rrr, int subs, int pixelsPerFrame);
void PhotoCompression (int pixelsPerFrame);
void DeltaStepwise (int *buf, int bufSizeInWords, int nbFrames);
void PhotoCompression2 (int pixelsPerFrame);

/*=================================================================================================*/
/*External function and variables
===================================================================================================*/

extern int vbwl_short (u4 *source, u4 words, u4 *dest, int blockSize);
extern int fm_ariencode (u4 *source, u4 nwords, u4 *dest, u4 *swap, int modelId);


/*
Transforms a buffer where data is stored frame by frame into a buffer where data is stored
pixels by pixels.

SRC contains : p0f0, p1f0, p2f0, ... pPf0, p0f1, p1f1, ...
DEST contains: p0f0, p0f1, p0f2, ... p0fF, p1f0, p1f1, p1f2, ...
*/
void Frameify (int *src, int srcSizeInWords, int pixelsPerFrame, int *dest)
{
    int i, j;

    int nbFrames = roldiv(srcSizeInWords, pixelsPerFrame);

    for (i=0; i < pixelsPerFrame; i++)
	{
		for (j=0; j < nbFrames; j++)
		{
	    	dest[j*pixelsPerFrame + i] = src[i*nbFrames + j];
		}    
	}
	return;
}


/*
The source buffer is ordered frame by frame.
The destination buffer contains the first frame (the keyframe) and then the differences to this keyframe.
*/
void Keyframe (int *src, int srcSizeInWords, int pixelsPerFrame, int *dest)
{
    int i, j;	 
    int nbFrames = roldiv(srcSizeInWords, pixelsPerFrame);
    
    // copy first frame
    for (i=0; i < pixelsPerFrame; i++)
	{
		dest[i] = src[i];
	}

    for (j=1; j < nbFrames; j++)
	{
		for (i=0; i<pixelsPerFrame; i++)
		{
			dest[i+j*pixelsPerFrame] = src[i+j*pixelsPerFrame] - src[i];
		}
	}
    
    return;
}

/*differences are made in place, from bottom to top, with 16 bit sign extension
Note: there is an assembly version of this function
*/
void _Delta (int *buf, int words)
{
    int i;

    for (i=words-1; i>0; i--)
	{ 
    	buf[i] = (buf[i] - buf[i-1]) &0xffff; 
		buf[i] = buf[i] > 0x7fff ? buf[i] | 0xffff0000 : buf[i];
	}

    return;
}

/*
Transforms a buffer where data is stored pixels by pixel into a buffer where data is stored frame by frame.

SRC contains: p0f0, p0f1, p0f2, ... p0fF, p1f0, p1f1, p1f2, ...
DEST contains : p0f0, p1f0, p2f0, ... pPf0, p0f1, p1f1, ...
*/
void Pixelize (int *src, int srcSizeInWords, int pixelsPerFrame, int *dest)
{
    int i, j;

    int nbFrames = roldiv(srcSizeInWords, pixelsPerFrame);

	for (i=0; i < nbFrames; i++)
	{
		for (j=0; j < pixelsPerFrame; j++)
		{
			dest[j*nbFrames + i] = src[i*pixelsPerFrame + j];
		}
	}

    return;
}


/*
This function transforms signed data to unsigned data with keeping the distribution around zero.
The source data is 'noise' around zero and we want to keep that in the unsigned data.
For this reason, positive numbers are multiplied by two and negative numbers are inserted in between.
*/
void Map2Pos32 (int *buf, int sizeInWords)
{
    int i;

    for (i=0; i < sizeInWords; i++)
	{
		buf[i] = buf[i] < 0 ? (buf[i] * -2 -1) & 0xffff : buf[i] * 2;
	}
    return;
}



/*
Temporal-spatial reduction for spectro:  The RampDiff algorithm.

input data order must be: p0f0 p0f1 p0f2 ... pixelized
*/
int RampDiff (u4 *buf, int size_w, int rrr, int subs, int pixelsPerFrame)
{
    int a, b;
    int i, j, k;
    int modrrr = rrr - DetectorConstantsSpectro.DiscardPre - DetectorConstantsSpectro.DiscardPost;
    int frames = roldiv(size_w, pixelsPerFrame);    
    int spr = roldiv(modrrr, subs);
    int ramps = roldiv(frames, spr); 
    int key;
	
#define DELTA(buf,a,b) buf[a] = (buf[a] - buf[b]) & 0xffff; buf[a] = buf[a] > 0x7fff ? buf[a] | 0xffff0000 : buf[a]
	
    // delta each ramp
    // count over ramps
    for (i = 0; i < size_w; i+=spr) // hop from ramp to ramp
	{
		for (j = spr-1; j > 0; j--) // delta each keyramp
		{ // this is the same strategy as in the "delta" function used in photometry
			a = i + j; b = a - 1;
			DELTA(buf,a,b);
		}
	}
	// now we are in a 16 bit signed system with signs extended to 32 bit
	// and we have to ensure we stay in that system until the map2pos function is called
	
	// keyramp
	// loop over pixels = buffer, increased with framesize
	for (i = 0; i < size_w; i+=frames)	
	{
		// loop over the subs per ramp = keyramp 
		for (j = 0; j < spr; j++)
		{
			// loop over the ramps (caveat: incomplete ramps are skipped)
			for (k = 1; k < ramps; k++)
			{
				a = i + j + k*spr; b = i + j;
				DELTA(buf,a,b);
			}
		}
	}
	// delta the keyramp
	// count over ramps
	for (i = 0; i < pixelsPerFrame; i++) // for each pixel we have a keyramp
	{			
		for (j = spr-1; j > 1; j--) // delta each keyramp
		{
			a = i*frames + j; b = a - 1;
			DELTA(buf,a,b);
		}
	}	
	
	// now calculate the keysample
	key = 0;
	for (i = 0; i < pixelsPerFrame; i++) // for each pixel we have a keyramp
	{
		key += buf[i*frames];
	}
	// frame 1
	key = roldiv(key, pixelsPerFrame);
	
	for (i = 0; i < pixelsPerFrame; i++)
	{ 
		a = i*frames;
		buf[a] = (buf[a] - key) & 0xffff; // this one is different, so we don't use DELTA()
		buf[a] = buf[a] > 0x7fff ? buf[a] | 0xffff0000 : buf[a];
	}
	
	return key;
}



/*
entrypoint for temporal/spatial lossless processing in photometry

This function implements the Scheme 0 (previously called 'velveteen')
*/
void PhotoCompression (int pixelsPerFrame)
{
	u4* source = ProcBuff;
	u4* swap = DramBuffer;

	// convert data from pixel order to frame order
	Frameify((int*)source+1, source[0], pixelsPerFrame, (int*)swap+1);
	swap[0] = source[0];
	
	// apply the keyframe (frames - first frame)
	Keyframe ((int*)swap+1, swap[0], pixelsPerFrame, (int*)source+1);
	
	// make differences
	Delta ((int*)source+1, source[0]);
	
	// transform data into pixel order
	Pixelize ((int*)source+1, source[0], pixelsPerFrame, (int*)swap+1);
	swap[0] = source[0];
	
	// also make differences here
	Delta ((int*)swap+1, swap[0]);
	
	// map to positive
    Map2Pos32 ((int*)swap+1, swap[0]);
	
	CompressedDataBuffer.Data = ProcBuff + 1;
    CompressedDataBuffer.SizeInWords = vbwl_short(swap+1, swap[0], CompressedDataBuffer.Data, 4);

	return;
}



/*
This function performs a kind of key-framing
*/
void DeltaStepwise (int *buf, int bufSizeInWords, int nbFrames)
{
    int i,j;
	
	j=0;
	for (i=bufSizeInWords-1, j=1; i>0; i--)
	{
		if (j < nbFrames)
		{
			buf[i] = (buf[i] - buf[i-1]) &0xffff;
			buf[i] = buf[i] > 0x7fff ? buf[i] | 0xffff0000 : buf[i];
			j++;
		}
		else
		{
			j = 1;
		}
	}
	
	return;
}

/*
entrypoint for temporal/spatial lossless processing in photometry

This function implements the Scheme 1 (previously called 'velvet')
*/
void PhotoCompression2 (int pixelsPerFrame)
{
	u4* src = ProcBuff;
	u4* swap = DramBuffer;
	u4 srcSizeInWords = src[0];
    int nbFrames = roldiv(srcSizeInWords, pixelsPerFrame);
 
	// kind of key-framing
    DeltaStepwise ((int*)src+1, srcSizeInWords, nbFrames);
    
    // transform from pixel order to frame order
	Frameify ((int*)src+1, srcSizeInWords, pixelsPerFrame, (int*)swap);

	// make differences
    Delta ((int*)swap, srcSizeInWords);

	// transform into pixel order again
    Pixelize ((int*)swap, srcSizeInWords, pixelsPerFrame, (int*)(SortedDataBuffer.Data));

	//map to positive numbers
	Map2Pos32 ((int*)(SortedDataBuffer.Data), srcSizeInWords);

	CompressedDataBuffer.Data = ProcBuff + 1;
    CompressedDataBuffer.SizeInWords = fm_ariencode (SortedDataBuffer.Data, srcSizeInWords, CompressedDataBuffer.Data, swap, PHOT_MODEL);
}
