/*****************************************************************************************
* File name: Average.c
* Purpose:  Performs the Reduction of the Photometric Readouts			 
*
* Description:	The Robust Averaging Program. It is used in Photometry Mode to
*				Calculate Samples Average.
******************************************************************************************/

#ifdef SIMULATOR
#include "virtuosoSim.h"
#else // SIMULATOR
#include "node1.h"
#include "k_struct.h"
#endif
#include "spuAsw.h"
#include "spulib.h"

/*=================================================================================================*/
/* Stuff needed for the random number supported rounding in StatAveraging */
#define	URAND_A_val 1664525
#define	URAND_C_val 32767
#define	URAND_CLIP 0x7fffffff
//#define	URAND_RMX 2147483647
#define URAND_RMXH 1073741823
static int URAND_seed = 1;
/*=================================================================================================*/

/*=================================================================================================*/
/*Functions declared in this file
===================================================================================================*/
int FloatAveraging (int *source, int nbPixelsPerFrame, int nbFrames, int *dest, int numavg);
int StatAveraging (int *source, int nbPixelsPerFrame, int nbFrames, int *dest, int numavg, int RoundingBits, int strategy);
int Averaging4Samples (int *source, int nbPixelsPerFrame, int nbFrames, int *dest);
int Averaging8Samples (int *source, int nbPixelsPerFrame, int nbFrames, int *dest);
static float StatRestKoryo (int average, float invden);
static float StatRestRoland (int average, float invden);
void Clamp (int *source, int nsamples, int RoundingBits);
void ClampSpec (int *source, int nsamples, int RoundingBits);

/* this is for Compressed Sensing */
// the size of a vector to be transformed
#define NOISELET_CHUNK 256
// the shift is the sqrt of the chunk in bits: 1024->32 -> 5 bits, 256->16 -> 4 bits
#define NOISELET_SHIFT 4
void NoiseletTransform1d (int *source, int size, int *dest);
void NoiseletNorm1d (int *source, int size, int rshift);
int MaskData (int *source, int size, int *dest, int from, int every);
int Csensing (int *source, int nPixels, int nFrames, int *dest, int ReductionFactor);
/*=================================================================================================*/

// source and dest can be identical
int FloatAveraging (int *source, int nbPixelsPerFrame, int nbFrames, int *dest, int numavg)
{
    int pctr, bctr, sctr; // counters for pixels, blocks, sample, frames
    int blocks = 0; // number of averages to generate
    int remain = 0; // remaining frames 
    int avg = 0;
    int destctr = 0;
    int offset;
    float favg, factor, rem_factor;

    blocks = roldiv(nbFrames, numavg);
    remain = nbFrames - blocks * numavg;

	factor = 1.0f / numavg;
	if (remain) // note: if we were rounding we would have to use rem_den here as in StatAveraging
		rem_factor = 1.0f / remain;
	else
		rem_factor = 1.0f;

    for (pctr=0; pctr < nbPixelsPerFrame; pctr++)
    {	
		for (bctr=0; bctr < blocks; bctr++)	
		{
	    	offset = pctr*nbFrames + bctr*numavg;
	    	for (sctr=0, avg=0; sctr < numavg; sctr++)	
			{
				avg += source[offset + sctr];
			}
	    	    
// FIXME: avg ist already 0 here when compiled with -O2

	    	favg = (float)avg * factor;

			dest[destctr++] = (int)(favg + 0.5);

		}
		// last block
		if (remain)
		{   
	    	offset = pctr*nbFrames + bctr*numavg;	    
	    	for (sctr=0, avg=0; sctr < remain; sctr++)
			{
				avg += source[offset + sctr];
			}

	    	favg = (float)avg * rem_factor;

			dest[destctr++] = (int)(favg + 0.5);
		}
    }
    
    return destctr;
}


// source and dest can be identical
int Averaging4Samples (int *source, int nbPixelsPerFrame, int nbFrames, int *dest)
{
    int pctr, bctr, sctr; // counters for pixels, blocks, sample
    int blocks = 0; // number of averages to generate
    int remain = 0; // remaining frames 
	int remain2;
    int avg = 0;
    int destctr = 0;
    int offset;
    //
    int numavg = 4;

    blocks = nbFrames >> 2;
    remain = nbFrames - blocks * numavg;
	remain2 = remain * (1 << DetectorConstantsPhoto.RoundingBits);

    for (pctr=0; pctr < nbPixelsPerFrame; pctr++)
    {
		for (bctr=0; bctr < blocks; bctr++)	
		{
			offset = pctr*nbFrames + bctr*numavg;
			avg = source[offset] + source[offset+1] + source[offset+2] + source[offset+3];
			dest[destctr++] = (avg + (RoundingHelper[DetectorConstantsPhoto.RoundingBits+2])) >> (2 + DetectorConstantsPhoto.RoundingBits); // round	    
		}
	
		// last block
		if (remain > 1)
		{   
	    	offset = pctr*nbFrames + bctr*numavg;	    
	    	for (sctr=0, avg=0; sctr < remain; sctr++)
			{
				avg += source[offset + sctr];
			}

			dest[destctr++] = roldivs(avg + ((remain2 + 1) >> 1), remain2);
		}

		if (remain == 1)
		{
	    	// no rounding needed
	    	dest[destctr++] = (source[pctr*nbFrames + bctr*numavg] + RoundingHelper[DetectorConstantsPhoto.RoundingBits]) >> DetectorConstantsPhoto.RoundingBits;
		}	
    }
    
    return destctr;
}

// source and dest can be identical
int Averaging8Samples (int *source, int nbPixelsPerFrame, int nbFrames, int *dest)
{
    int pctr, bctr, sctr; // counters for pixels, blocks, sample
    int blocks = 0; // number of averages to generate
    int remain = 0; // remaining frames 
    int avg = 0;
    int destctr = 0;
    int offset;
    //
    int numavg = 8;
	
    blocks = roldiv(nbFrames, numavg);
    remain = nbFrames - blocks * numavg;
	
    for (pctr=0; pctr < nbPixelsPerFrame; pctr++)
    {
		for (bctr=0; bctr < blocks; bctr++)	
		{
			offset = pctr*nbFrames + bctr*numavg;
			
			avg = source[offset] + source[offset+1] + source[offset+2] + source[offset+3] 
				+ source[offset+4] + source[offset+5] + source[offset+6] + source[offset+7];
			
			dest[destctr++] = (avg + 4) >> 3; // round	    
		}
		// last block
		if (remain > 1)
		{   
			offset = pctr*nbFrames + bctr*numavg;	    
			for (sctr=0, avg=0; sctr < remain; sctr++)
			{
				avg += source[offset + sctr];
			}
			
			dest[destctr++] = roldivs(avg + (remain>>1), remain);
		}
		if (remain == 1)
		{
			// no rounding needed
			dest[destctr++] = source[pctr*nbFrames + bctr*numavg];
		}	
    }
    
    return destctr;
}

static float StatRestKoryo (int average, float invden)
{
    int temp;
    float stat = 0.5;
    
    if (average < 0)
		stat = -0.5;
    
    URAND_seed = URAND_seed * URAND_A_val + URAND_C_val;
    temp = (URAND_seed & URAND_CLIP);
    if (temp > URAND_RMXH) 
		stat = stat - invden;
    
    return stat;
}

static float StatRestRoland (int average, float invden)
{
    int temp;
    float sign = 1.0;
    float stat = 0.5;
    
    if (average < 0)
	{
		sign = -1.0;
		stat = -0.5;
    }
    URAND_seed = URAND_seed * URAND_A_val + URAND_C_val;
    temp = (URAND_seed & URAND_CLIP);
    if (temp > URAND_RMXH) 
		stat = stat - sign*invden;
    
    return stat;
}

int StatAveraging (int *source, int nbPixelsPerFrame, int nbFrames, int *dest, int numavg, int RoundingBits, int strategy)
{
    int pctr=0, bctr=0, sctr=0; // counters for pixels, blocks, sample, frames
    int blocks = 0; // number of averages to generate
    int remain = 0; // remaining frames
    int avg = 0;
    int destctr = 0;
    int offset;
    int den;
    int rem_den;
    float favg, deninv, rem_deninv;    

	// choose between two different kinds of rounding.
	float (*StatRest) (int average, float invden) = StatRestRoland;
	if (strategy == K_ROUNDING_STYLE_KORYO)
		StatRest = StatRestKoryo;

    blocks = roldiv (nbFrames, numavg);
    remain = nbFrames - blocks * numavg;
    
    den = numavg * (1 << RoundingBits);
    deninv = 1.0f / (float)den; 

    if (remain)
    {
		rem_den = remain * (1 << RoundingBits);
    	rem_deninv = 1.0f / (float)rem_den; 
	}
    else
        rem_deninv = 1.0f; 

    for (pctr=0; pctr < nbPixelsPerFrame; pctr++)
    {
    	avg = 0;
        for (bctr=0; bctr < blocks; bctr++)
        {
            offset = pctr*nbFrames + bctr*numavg;
	    	avg = 0;
	    	sctr = 0;
            while (sctr < numavg)
	    	{	
				avg += source[offset + sctr];
				sctr++;                
	    	}
	    
	    	favg = (float)avg*deninv + StatRest(avg, deninv); // round towards +
			
	    	dest[destctr] = (int)favg;
	    	destctr++;
        }
        // last block
        if (remain)
        {
            offset = pctr*nbFrames + bctr*numavg;
	    	avg = 0;
	    	sctr = 0;
            while (sctr < remain)
            {
				avg += source[offset + sctr];
				sctr++;
	    	}

	    	favg = (float)avg*rem_deninv + StatRest(avg, rem_deninv); // round towards +

	    	dest[destctr] = (int)favg;
	    	destctr++;
        }
    }

    return destctr;
}

void Clamp (int *source, int nsamples, int RoundingBits)
{
	int ctr;
	int maxval = 32767, minval = -32768;

	if (RoundingBits == 1)
	{
		maxval = 16383;
		minval = -16384;
	}
	if (RoundingBits == 2)
	{
		maxval = 8191; 
		minval = -8192;
	}
	if (RoundingBits == 3)
	{
		maxval = 4095; 
		minval = -4096;
	}
	for (ctr=0; ctr < nsamples; ctr++)
	{
		if (source[ctr] > maxval)
			source[ctr] = maxval;
		if (source[ctr] < minval)
			source[ctr] = minval;
	}
	return;
}

void ClampSpec (int *source, int nsamples, int RoundingBits)
{
	int ctr;
	int maxval = 65535, minval = 0;

	if (RoundingBits == 1)
		maxval = 32767;
	
	if (RoundingBits == 2)
		maxval = 16383; 
	
	if (RoundingBits == 3)
		maxval = 8191; 

	for (ctr=0; ctr < nsamples; ctr++)
	{
		if (source[ctr] > maxval)
			source[ctr] = maxval;
		if (source[ctr] < minval)
			source[ctr] = minval;
	}
	return;
}

void NoiseletTransform1d (int *source, int size, int *dest)
{
  int c, d, j, k;
  int temp;

  c = size - 1;
  c = size >> 1; // non-dragon noiselets 
  
  for (j = 0; j < size >> 1; j++) 
  {
      k = j ^ c;
      dest[j] = source[j] + source[k];
      dest[k] = source[j] - source[k];
  }

  for (d = c >> 1; d > 0; d >>= 1) 
  {   
    for (j = 0; j < size >> 1; j++) 
    {
		k = j ^ c ^ d;
		temp = dest[j];
		dest[j] = dest[j] - dest[k];
		dest[k] = temp + dest[k];
    }
  }
  return;
}

void NoiseletNorm1d (int *source, int size, int rshift)
{
    int i;
    
    for (i=0; i<size; i++)
		source[i] >>= rshift;
    
    return;
}

int MaskData (int *source, int size, int *dest, int from, int every) // works in place
{
    int i;
    int j = 0;
    for (i = from; i < size; i += every)
		dest[j++] = source[i];

    return j;
}

int Csensing (int *source, int nPixels, int nFrames, int *dest, int ReductionFactor)
{
	int chunk; // counter over all chunks
	int ChunksPerFrame = roldiv (nPixels, NOISELET_CHUNK);
	int nChunks = nFrames * ChunksPerFrame;
	int destoffset = 0;	
	int shift = 0;

	for (chunk=0; chunk < nChunks; chunk++) // every chunk
	{
		// transform each chunk
		NoiseletTransform1d (source+chunk*NOISELET_CHUNK, NOISELET_CHUNK, dest+destoffset);

		// normalize
		NoiseletNorm1d (dest+destoffset, NOISELET_CHUNK, NOISELET_SHIFT);

		// mask frame and collapse in place
		shift = (roldiv (chunk, ChunksPerFrame) % ReductionFactor);
		destoffset += MaskData (dest+destoffset, NOISELET_CHUNK, dest+destoffset, shift, ReductionFactor);
	}

	return destoffset;
}