/******************************************************************************************
* File name: Bol_ex.c
* Purpose: Performs Default Data Compression for Photometry.
*
* Description:	This Program Performs Data Compression in Photometry for PhotoReductionAndCompression
*				Modes Depending on the Received Compression Parameter.
******************************************************************************************/

#ifdef SIMULATOR
	#include "virtuosoSim.h"
	#include "stdio.h"
#else // SIMULATOR
	#include "node1.h"
	#include "k_struct.h"
#endif

#include "spuAsw.h"
#include "spulib.h"

/*=================================================================================================*/
/*Functions declared in this file
===================================================================================================*/

void _ExpandSign(int *buf, int sizeInWords);
void _MakeDataSigned(int *buf, int sizeInWords);
void PhotoReductionAndCompression(u4 numavg, u4 nbFrames, COMPRESSION_MODE mode);

/*=================================================================================================*/
/*External function and variables
===================================================================================================*/

extern void PhotoCompression(int pixelsPerFrame);
extern void PhotoCompression2(int pixelsPerFrame);

extern int vbwl_short (u4 *source, u4 words, u4 *dest, int blockSize);

extern int Averaging4Samples (int *source, int nbPixelsPerFrame, int nbFrames, int *dest);
extern int Averaging8Samples (int *source, int nbPixelsPerFrame, int nbFrames, int *dest);
extern int FloatAveraging (int *source, int nbPixelsPerFrame, int nbFrames, int *dest, int numavg);
extern int StatAveraging (int *source, int nbPixelsPerFrame, int nbFrames, int *dest, int numavg, int RoundingBits, int strategy);
extern int DecimatePixelOrder (int *buffer, u4 framesize, u4 frames, u4 groupsize, u4 discard_pre, u4 discard_post);
void Clamp (int *source, int nsamples, int RoundingBits);
int Csensing (int *source, int nPixels, int nFrames, int *dest, int ReductionFactor);
extern u4 AveragingParametersSummary;
extern int ReductionStatus;
/****************************************************************************************
* Function name: ExpandSign
* Purpose: The input buffer contains 1 signed short per word.
*          The sign of the short is extended on the whole word
*          This function works in place
* Note: there is an assembly version of this function
*****************************************************************************************/

void _ExpandSign(int *buf, int sizeInWords)
{
    int i;

    for (i=0; i<sizeInWords; i++)
	{
		buf[i] = buf[i] > 0x7fff ? buf[i] | 0xffff0000 : buf[i];
	}
    
    return;
}


/****************************************************************************************
* Function name: MakeDataSigned
* Purpose: The input buffer contains 1 unsigned short per word.
*          The values are translated from [0, 65535] to [-32768, 32767].
*          In this operation, the sign is extended on the whole word.
*          This function works in place
* Note: there is an assembly version of this function
*****************************************************************************************/
void _MakeDataSigned(int *buf, int sizeInWords)
{
	int i;

	for (i=0; i<sizeInWords; i++)
	{
		buf[i] = buf[i] - 0x8000;
	}

	return;
}


void PhotoReductionAndCompression(u4 numavg, u4 nbFrames, COMPRESSION_MODE mode)
{
	u4 dataBufferSize = nbFrames * K_NB_PIXELS_IN_PHOTO_ARRAY;

	/*=============================================================================================*/
	/* Call Sequence of Compression Tasks
	===============================================================================================*/

	/* DMC Header Extraction and Compression */
	DmcHeaderCompression(nbFrames, K_PHOTO_HEADER_SIZE, mode);

	if(mode.DataBlockId!=5) /* Blue Bolometers */
	{
		/* Be Sure That Number of Raw Channels is Compatible with Number of selected pixels*/
		if (RawChannelSelectionPhoto.NbSelectedRawChannels > DxsPhotoBlue.NbSelectedPixels)
		{
			RawChannelSelectionPhoto.NbSelectedRawChannels = DxsPhotoBlue.NbSelectedPixels;
		}

		/* Perform the Compression of the Science data for the 4 Sub-Arrays*/
		/* Detector Selection */

		Dxs(dataBufferSize, &DxsPhotoBlue, nbFrames, K_CBUF_SCIENCE_BLUE_PHOT1, 0, mode);

		/* preprocessing of the data */
		PreProcessing(mode.DetectorType);

		// SIZE will be stored in [0] and is given in words
		// make data signed int if they aren't and expand the sign

		// signed short data -> expand sign to int; unsigned data will be mapped to signed int
		if (mode.BolcDataType == 0)	/*DataType = S-R*/
		{
			ExpandSign((int*)ProcBuff+1, ProcBuff[0]); // signed int -> only expand the sign
		}
		else
		{
			MakeDataSigned((int*)ProcBuff+1, ProcBuff[0]);	// transpose from 0..65535 to -32768..32767	
		}

		// here the REDUCTION starts
		ReductionStatus = 0; // for hk
		if (mode.CompressionMode == K_CM_PHOTO_COMPRESSED_SENSING)
		{
			ProcBuff[0] = Csensing ((int*)ProcBuff+1, DxsPhotoBlue.NbSelectedPixels, nbFrames, (int*)ProcBuff+1, 6);
			Result.ReductionAlgo = 6;
			AveragingParametersSummary = 0xd0000000 | numavg << 16;
		}
		else
		{
			// averaging
			// choose 13.8 hardcoded averaging lengths (ignore override) and the traditional algorithms
			if (DetectorConstantsPhoto.RoundingStrategy == K_ROUNDING_STYLE_138)
			{
				AveragingParametersSummary = 0xd0000000 | numavg << 16;
				
				if (numavg == 8) // default double mode, no rounding
				{
					ProcBuff[0] = Averaging8Samples ((int*)ProcBuff+1, DxsPhotoBlue.NbSelectedPixels, nbFrames, (int*)ProcBuff+1);
					Result.ReductionAlgo = 8;
				}
				else if (numavg == 4) // default averaging of 4, with rounding
				{
					ProcBuff[0] = Averaging4Samples ((int*)ProcBuff+1, DxsPhotoBlue.NbSelectedPixels, nbFrames, (int*)ProcBuff+1);
					Result.ReductionAlgo = 4;
					AveragingParametersSummary |= (DetectorConstantsPhoto.RoundingBits & 0x3) << 26;  //the rounding has been done in the averaging step
				}
				else // half compression mode, float alg, no rounding
				{
					ProcBuff[0] = FloatAveraging ((int*)ProcBuff+1, DxsPhotoBlue.NbSelectedPixels, nbFrames, (int*)ProcBuff+1, numavg);
					Result.ReductionAlgo = 0x80 | numavg;
				}
			}
			else // choose new fully flexible and statistically correct algorithm
			{
				// do we have an override of the number of samples to average?
				if (DetectorConstantsPhoto.NbSamplesToAverage != 0)
					numavg = DetectorConstantsPhoto.NbSamplesToAverage;

				/* new reduction step in version 13.10: frame dropping, needs to modify numavg */
				nbFrames = DecimatePixelOrder ((int*)ProcBuff+1, DxsPhotoBlue.NbSelectedPixels, nbFrames, numavg, DetectorConstantsPhoto.DiscardPre, DetectorConstantsPhoto.DiscardPost);
				ProcBuff[0] = nbFrames * DxsPhotoBlue.NbSelectedPixels; // update buffer size
				if ((DetectorConstantsPhoto.DiscardPre + DetectorConstantsPhoto.DiscardPost) < numavg)
					numavg -= (DetectorConstantsPhoto.DiscardPre + DetectorConstantsPhoto.DiscardPost);

				Result.ReductionAlgo = 0x80 | numavg; // use float identifier
				AveragingParametersSummary = 0xd0000000 | numavg << 16;

				// now choose strategy
				if (DetectorConstantsPhoto.RoundingStrategy == K_ROUNDING_STYLE_ROLAND)
					ProcBuff[0] = StatAveraging ((int*)ProcBuff+1, DxsPhotoBlue.NbSelectedPixels, nbFrames, (int*)ProcBuff+1, numavg, DetectorConstantsPhoto.RoundingBits, K_ROUNDING_STYLE_ROLAND);
				
				else // Koryo style
					ProcBuff[0] = StatAveraging ((int*)ProcBuff+1, DxsPhotoBlue.NbSelectedPixels, nbFrames, (int*)ProcBuff+1, numavg, DetectorConstantsPhoto.RoundingBits, K_ROUNDING_STYLE_KORYO);
				
				// clamp the averaged values to 16 bit
				Clamp ((int*)ProcBuff+1, ProcBuff[0], DetectorConstantsPhoto.RoundingBits);
				// insert rounding information
				AveragingParametersSummary |= (DetectorConstantsPhoto.RoundingBits & 0x3) << 26;

				/* for 13.10 the hi-byte is used for decimation and sync delay instead of rounding */
				/* the rounding info is anyway transported via the NULLWORD (compressedEntityBuffer[2]) */
				/* the HK will overwrite this hi-byte with its own info */
				// Result.ReductionAlgo |= (AveragingParametersSummary & 0xFF000000) >> 16;
				Result.ReductionAlgo |= (DetectorConstantsPhoto.DiscardPre & 0x07) << 13;
				Result.ReductionAlgo |= (DetectorConstantsPhoto.DiscardPost & 0x07) << 10;
				Result.ReductionAlgo |= (DetectorConstantsPhoto.SyncDelay & 0x03) << 8;
				ReductionStatus = DetectorConstantsPhoto.RoundingBits;
			}
		}
		/* Temporal + Spatial Redundancy Reduction and lossless encoding */

		if (CompressionSoftwareRunning == K_YES) //check that the compression software has not been stopped in the meanwhile
		{
			if (mode.CompressionMode != K_CM_PHOTO_COMPRESSED_SENSING)
			{
				if (DetectorConstantsPhoto.Scheme == 0) // select compression scheme
				{
					// velveteen
					PhotoCompression(DxsPhotoBlue.NbSelectedPixels);
					// enter key: not necessary, because it is already 0, d0 stays d0
				}
				else
				{
					// velvet
					PhotoCompression2(DxsPhotoBlue.NbSelectedPixels);
					// enter key
					AveragingParametersSummary |= 0x01000000; // d0 now becomes d1
#ifdef SIMULATOR
					FILE* fp = fopen("cd.txt", "wt+");
					for (unsigned int i=0; i < ProcBuff[0]; i++)
					{
						fprintf(fp, "%08x\n", ProcBuff[i+1]);
					}
					fclose(fp);
#endif// SIMULATOR
				}
			}
			else // LLC scheme for Compressed Sensing
			{
				PhotoCompression(DxsPhotoBlue.NbSelectedPixels);
			}

			/*Look now if Raw Channels were selected for further llc */
			if(RawChannels.SizeInWords)
			{
				CompressRawChannels(&RawChannels, mode.DetectorType);
			}

			/* Write the Compressed Data in the Output Buffer */
			BuildCompressedEntity(mode);
		}
	}
	else /* Red Bolometers */
	{
		/* Be Sure That Number of RC is Compatible with Number of DX*/
		if (RawChannelSelectionPhoto.NbSelectedRawChannels > DxsPhotoRed.NbSelectedPixels)
		{
			RawChannelSelectionPhoto.NbSelectedRawChannels = DxsPhotoRed.NbSelectedPixels;
		}

		/* Detector Selection */
		Dxs(dataBufferSize, &DxsPhotoRed, nbFrames, K_CBUF_SCIENCE_RED_PHOT, 0, mode);

		/* preprocessing of the data */
 		PreProcessing(mode.DetectorType);

		// make data signed int if they aren't and expand the sign

		if (mode.BolcDataType == 0)	/*DataType = S-R*/
		{
			ExpandSign((int*)ProcBuff+1, ProcBuff[0]); // signed int -> only expand the sign
		}
		else
		{
			MakeDataSigned((int*)ProcBuff+1, ProcBuff[0]);	// transpose from 0..65535 to -32768..32767
		}

		// here the REDUCTION starts
		ReductionStatus = 0; // for hk
		if (mode.CompressionMode == K_CM_PHOTO_COMPRESSED_SENSING)
		{
			ProcBuff[0] = Csensing ((int*)ProcBuff+1, DxsPhotoRed.NbSelectedPixels, nbFrames, (int*)ProcBuff+1, 6);
			Result.ReductionAlgo = 6;
			AveragingParametersSummary = 0xd0000000 | numavg << 16;
		}
		else
		{
			// averaging
			// choose 13.8 hardcoded averaging lengths (ignore override) and the traditional algorithms
			if (DetectorConstantsPhoto.RoundingStrategy == K_ROUNDING_STYLE_138)
			{
				AveragingParametersSummary = 0xd0000000 | numavg << 16;

				if (numavg == 8) // default double mode, no rounding
				{
					ProcBuff[0] = Averaging8Samples ((int*)ProcBuff+1, DxsPhotoRed.NbSelectedPixels, nbFrames, (int*)ProcBuff+1);
					Result.ReductionAlgo = 8;
				}
				else if (numavg == 4) // default averaging of 4, with rounding
				{
					ProcBuff[0] = Averaging4Samples ((int*)ProcBuff+1, DxsPhotoRed.NbSelectedPixels, nbFrames, (int*)ProcBuff+1);
					Result.ReductionAlgo = 4;
					AveragingParametersSummary |= (DetectorConstantsPhoto.RoundingBits & 0x3) << 26;  //the rounding has been done in the averaging step
				}
				else // half compression mode, float alg, no rounding
				{
					ProcBuff[0] = FloatAveraging ((int*)ProcBuff+1, DxsPhotoRed.NbSelectedPixels, nbFrames, (int*)ProcBuff+1, numavg);
					Result.ReductionAlgo = 0x80 | numavg;
				}
			}
			else // choose new fully flexible and statistically correct algorithm
			{
				// do we have an override of the number of samples to average?
				if (DetectorConstantsPhoto.NbSamplesToAverage != 0)
					numavg = DetectorConstantsPhoto.NbSamplesToAverage;

				/* new reduction step in version 13.10: frame dropping, needs to modify numavg */
				nbFrames = DecimatePixelOrder ((int*)ProcBuff+1, DxsPhotoRed.NbSelectedPixels, nbFrames, numavg, DetectorConstantsPhoto.DiscardPre, DetectorConstantsPhoto.DiscardPost);
				ProcBuff[0] = nbFrames * DxsPhotoRed.NbSelectedPixels; // update buffer size
				if ((DetectorConstantsPhoto.DiscardPre + DetectorConstantsPhoto.DiscardPost) < numavg)
					numavg -= (DetectorConstantsPhoto.DiscardPre + DetectorConstantsPhoto.DiscardPost);

				Result.ReductionAlgo = 0x80 | numavg; // use float identifier
				AveragingParametersSummary = 0xd0000000 | numavg << 16;

				// now choose strategy
				if (DetectorConstantsPhoto.RoundingStrategy == K_ROUNDING_STYLE_ROLAND)
					ProcBuff[0] = StatAveraging ((int*)ProcBuff+1, DxsPhotoRed.NbSelectedPixels, nbFrames, (int*)ProcBuff+1, numavg, DetectorConstantsPhoto.RoundingBits, K_ROUNDING_STYLE_ROLAND);
				else // Koryo style
					ProcBuff[0] = StatAveraging ((int*)ProcBuff+1, DxsPhotoRed.NbSelectedPixels, nbFrames, (int*)ProcBuff+1, numavg, DetectorConstantsPhoto.RoundingBits, K_ROUNDING_STYLE_KORYO);

				// clamp the averaged values to 16 bit
				Clamp ((int*)ProcBuff+1, ProcBuff[0], DetectorConstantsPhoto.RoundingBits);
				// insert rounding information
				AveragingParametersSummary |= (DetectorConstantsPhoto.RoundingBits & 0x3) << 26;

				/* for 13.10 the hi-byte is used for decimation and sync delay instead of rounding */
				/* the rounding info is anyway transported via the NULLWORD (compressedEntityBuffer[2]) */
				/* the HK will overwrite this hi-byte with its own info */
				// Result.ReductionAlgo |= (AveragingParametersSummary & 0xFF000000) >> 16;
				Result.ReductionAlgo |= (DetectorConstantsPhoto.DiscardPre & 0x07) << 13;
				Result.ReductionAlgo |= (DetectorConstantsPhoto.DiscardPost & 0x07) << 10;
				Result.ReductionAlgo |= (DetectorConstantsPhoto.SyncDelay & 0x03) << 8;
				ReductionStatus = DetectorConstantsPhoto.RoundingBits;
			}
		}
		/* Temporal + Spatial Redundancy Reduction and lossless encoding */
		if (CompressionSoftwareRunning == K_YES) //check that the compression software has not been stopped in the meanwhile
		{
			if (mode.CompressionMode != K_CM_PHOTO_COMPRESSED_SENSING) // normally...
			{
				if (DetectorConstantsPhoto.Scheme == 0) // select compression scheme
				{
					// velveteen
					PhotoCompression(DxsPhotoRed.NbSelectedPixels);
					// enter key: not necessary, because it is already 0, d0 stays d0
				}
				else
				{
					// velvet
					PhotoCompression2(DxsPhotoRed.NbSelectedPixels);
					// enter key
					AveragingParametersSummary |= 0x01000000; // d0 now becomes d1
				}
			}
			else // LLC scheme for Compressed Sensing
			{
				PhotoCompression(DxsPhotoRed.NbSelectedPixels);
			}

			/* Look now if Raw Channels were selected for further llc */
			if(RawChannels.SizeInWords != 0)
			{
				CompressRawChannels(&RawChannels, mode.DetectorType);
			}

			/* Write the Compressed Data in the Output Buffer*/
			BuildCompressedEntity(mode);
		}
	}
}
