/********************************************************************************
* File name: fill_in.c
* Purpose:  Handles the Circular Buffers.
*
********************************************************************************/

#ifdef SIMULATOR
#include "virtuosoSim.h"
#else // SIMULATOR
#include "node1.h"
#include "k_struct.h"
#endif
#include "spuAsw.h"
#include "spulib.h"


/*=================================================================================================*/
/*Functions declared in this file (circular buffer functions)
===================================================================================================*/

void DeclareCircularBuffer(u4 bufferId, u4* start, u4* end);
u4 WriteInCircularBuffer(u4 bufferId, u4* src, u4 size);
u4 CircularBufferQuerySpace(u4 bufferId);
u4 CircularBufferQueryUsed(u4 bufferId);
u4 ReadInCircularBuffer(u4 bufferId, u4* dest, u4 size);
u4 ReadInCircularBufferAndPut2ShortsIn2Words(u4 bufferId, u4* dest, u4 size, u4 shortsAreSigned);
void CircularBufferRemoveLastWrittenValues(u4 bufferId, u4 size);
void CircularBufferRemove(u4 bufferId, u4 size);

/*=================================================================================================*/
/*Global variables declared in this file
===================================================================================================*/
 
u4*	CbufStart[K_CBUF_NB_CIRCULAR_BUFFERS];
u4	CbufSize[K_CBUF_NB_CIRCULAR_BUFFERS];
u4	CbufReadPos[K_CBUF_NB_CIRCULAR_BUFFERS];
u4	CbufWritePos[K_CBUF_NB_CIRCULAR_BUFFERS];

#ifdef SIMULATOR
u4 DmcHeaderCircularBuffer[K_CIRC_BUFFER_END_ADDR_DMCH], ScienceCircularBuffer[K_CIRC_BUFFER_END_ADDR + 0xFDFF];
#endif

/*=================================================================================================*/
/*External function and variables
===================================================================================================*/

extern u4 DmcPacketCounter;							//counts the frame received to know when we need to start compressing a buffer
extern u4 DmcHeaderStored;							//counts the number of header that have been stored since we started buffering
extern u4 DmcReadoutsStored[6];						//counts the number of readouts received since we started buffering (this is used when erasing circular buffers).  0 is spectro, 1 to 5 is photo (according to block id)


/****************************************************************************************
* Function name: ResetCircularBuffers
* Purpose: Declares the Circular Buffers and Clean the Tables and Parameters related to science data, dmc header and telemetry
*****************************************************************************************/
void ResetCircularBuffers()
{
	int i=0;
	
	/*=============================================================================================*/
	/* Reset the Global Variables to Initial Values */    
	
	DmcPacketCounter = 0;
	DmcHeaderStored = 0;
	for (i=0; i < 6; i++)
		DmcReadoutsStored[i] = 0;
	StoreNow = K_NO;
	ProcBuff[0] = 0;
	RawChannels.SizeInWords = 0;
	
	/* Declare the Start and End Addresses of the Circular Buffers in DRAM 
	=======================================================================*/
#ifndef SIMULATOR
	/* Reset Memory Counter for Scrubbing Purpose */

	DeclareCircularBuffer(K_CBUF_HEADER, (u4*) K_CIRC_BUFFER_START_ADDR_DMCH,(u4*) K_CIRC_BUFFER_END_ADDR_DMCH);
	DeclareCircularBuffer(K_CBUF_SCIENCE_BLUE_PHOT1, (u4*) K_CIRC_BUFFER_START_ADDR,(u4*) K_CIRC_BUFFER_END_ADDR1);
	DeclareCircularBuffer(K_CBUF_SCIENCE_BLUE_PHOT2, (u4*) K_CIRC_BUFFER_END_ADDR1+4,(u4*) K_CIRC_BUFFER_END_ADDR2);
	DeclareCircularBuffer(K_CBUF_SCIENCE_BLUE_PHOT3, (u4*) K_CIRC_BUFFER_END_ADDR2+4,(u4*) K_CIRC_BUFFER_END_ADDR3);
	DeclareCircularBuffer(K_CBUF_SCIENCE_BLUE_PHOT4, (u4*) K_CIRC_BUFFER_END_ADDR3+4,(u4*) K_CIRC_BUFFER_END_ADDR);
	DeclareCircularBuffer(K_CBUF_SCIENCE_RED_PHOT, (u4*) K_CIRC_BUFFER_START_ADDR, (u4*) K_CIRC_BUFFER_END_ADDR);
	DeclareCircularBuffer(K_CBUF_TM, (u4*) K_CIRC_BUFFER_START_ADDR_TM, (u4*) K_CIRC_BUFFER_END_ADDR_TM);


#else //Under Windows

	DeclareCircularBuffer(K_CBUF_HEADER, DmcHeaderCircularBuffer + K_CIRC_BUFFER_START_ADDR_DMCH, DmcHeaderCircularBuffer + K_CIRC_BUFFER_END_ADDR_DMCH);
	DeclareCircularBuffer(K_CBUF_SCIENCE_BLUE_PHOT1, ScienceCircularBuffer + K_CIRC_BUFFER_START_ADDR, ScienceCircularBuffer + K_CIRC_BUFFER_END_ADDR1);
	DeclareCircularBuffer(K_CBUF_SCIENCE_BLUE_PHOT2, ScienceCircularBuffer + K_CIRC_BUFFER_END_ADDR1 + 4, ScienceCircularBuffer + K_CIRC_BUFFER_END_ADDR2);
	DeclareCircularBuffer(K_CBUF_SCIENCE_BLUE_PHOT3, ScienceCircularBuffer + K_CIRC_BUFFER_END_ADDR2 + 4, ScienceCircularBuffer + K_CIRC_BUFFER_END_ADDR3);
	DeclareCircularBuffer(K_CBUF_SCIENCE_BLUE_PHOT4, ScienceCircularBuffer + K_CIRC_BUFFER_END_ADDR3 + 4, ScienceCircularBuffer + K_CIRC_BUFFER_END_ADDR);
	DeclareCircularBuffer(K_CBUF_SCIENCE_RED_PHOT, ScienceCircularBuffer + K_CIRC_BUFFER_START_ADDR, ScienceCircularBuffer + K_CIRC_BUFFER_END_ADDR);
	DeclareCircularBuffer(K_CBUF_TM, ScienceCircularBuffer + K_CIRC_BUFFER_END_ADDR, ScienceCircularBuffer + K_CIRC_BUFFER_END_ADDR + 0xFDFF);

#endif
}


/****************************************************************************************
* Function name: DeclareCircularBuffer
* Purpose: Declares the Circular Buffers and clean it (fill it with zeros)
*****************************************************************************************/
void DeclareCircularBuffer(u4 bufferId, u4* start, u4* end)
{
	register u4 i;

	/*=============================================================================================*/
	/* Initialization of the Adresses */
	CbufStart[bufferId] = start;
	CbufSize[bufferId] = (u4) (end-start);
	CbufReadPos[bufferId] = 0;
	CbufWritePos[bufferId] = 0;

	/* Clean this Memory */
	for(i=0; i<CbufSize[bufferId]; i++) 
	{
		*(start+i) = 0;
	}
}

/****************************************************************************************
* Function name: WriteInCircularBuffer
* Purpose: writes the content of 'src' in a circular buffer
*****************************************************************************************/
u4 WriteInCircularBuffer(u4 bufferId, u4* src, u4 size)
{
	register u4 i,j;
	u4 *position;	
  
	/*=============================================================================================*/
	/* Check if Enough Space */
  
	if (CircularBufferQuerySpace(bufferId) > size) 
    {
		if ((CbufSize[bufferId]-CbufWritePos[bufferId]) >= size) /* is enough space before the end?*/
		{
			position = CbufStart[bufferId] + CbufWritePos[bufferId];
			for(i=0; i<size; i++)
			{
				*(position + i) = *(src +i);
			}
			CbufWritePos[bufferId] += i; /* i started with 0, but entered a value */
			return i;
		}
		else
		{
			j = WriteInCircularBuffer(bufferId, src, (CbufSize[bufferId] - CbufWritePos[bufferId])); // write to the end
			i = j; // the i from the recursion is stacked, so this is allowed
			CbufWritePos[bufferId] = 0; // hard reset
			j = WriteInCircularBuffer(bufferId, src + j, (size - j)); // write the rest from the beginning
			i += j;
			CbufWritePos[bufferId] = j; // hard set 
			return i;
		}
    }
	else
	{
		return 0;  
	}
}

/****************************************************************************************
* Function name: ReadInCircularBuffer
* Purpose: reads the content of a circular buffer and store it in 'dest'
*****************************************************************************************/
u4 ReadInCircularBuffer(u4 bufferId, u4* dest, u4 size)
{
	register u4 i,j;
	u4 *position;
	
	/*=============================================================================================*/
    /* Check if Size of Data Already Written Corresponds to the Required Dump Size */  
	if (CircularBufferQueryUsed(bufferId) >= size)
    {
		if ((CbufSize[bufferId] - CbufReadPos[bufferId]) >= size) 
		{
			position = CbufStart[bufferId] + CbufReadPos[bufferId]; 
			for (i=0; i<size; i++)
			{
				*(dest + i) = *(position + i);
			}
			CbufReadPos[bufferId] += i;
			return i;
		}
		else
		{
			j = ReadInCircularBuffer(bufferId, dest, (CbufSize[bufferId] - CbufReadPos[bufferId])); // read to the end
			i = j; // the i from the recursion is stacked, so this is allowed
			CbufReadPos[bufferId] = 0; // hard reset
			j = ReadInCircularBuffer(bufferId, dest + j, (size - j)); // read the rest from the beginning
			i += j;
			CbufReadPos[bufferId] = j; // hard set 
			return i;
		}
    }
	else
	{
		return 0;
	}
}





/****************************************************************************************
* Function name: ReadInCircularBufferAndPut2ShortsIn2Words
* Purpose: reads the content of a circular buffer and store it in 'dest'.
*          It is assumed that the circular buffers contain 2 shorts per word.
*          The 2 shorts are stored in two separate words in 'dest'
*          If the shorts are signed, it has to be notified to the function such that
*          the sign can be expanded to the 32bits.
*****************************************************************************************/
u4 ReadInCircularBufferAndPut2ShortsIn2Words(u4 bufferId, u4* dest, u4 size, u4 shortsAreSigned)
{
	register u4 i,j=0;
	u4 *position;
	
	/*=============================================================================================*/
    /* Check if Size of Data Already Written Corresponds to the Required Dump Size */  
	if (CircularBufferQueryUsed(bufferId) >= size)
    {
		if ((CbufSize[bufferId] - CbufReadPos[bufferId]) >= size) 
		{
			position = CbufStart[bufferId] + CbufReadPos[bufferId]; 

			if (shortsAreSigned)
			{
				Put2ShortsIn2WordsSigned(position, dest, size);
			}
			else
			{
				Put2ShortsIn2Words(position, dest, size);
			}

			CbufReadPos[bufferId] += size;
			return size;
		}
		else
		{
			j = ReadInCircularBufferAndPut2ShortsIn2Words(bufferId, dest, (CbufSize[bufferId] - CbufReadPos[bufferId]), shortsAreSigned); // read to the end
			i = j; // the i from the recursion is stacked, so this is allowed
			CbufReadPos[bufferId] = 0; // hard reset
			j = ReadInCircularBufferAndPut2ShortsIn2Words(bufferId, dest + (j*2), (size - j), shortsAreSigned); // read the rest from the beginning
			i += j;
			CbufReadPos[bufferId] = j; // hard set 
			return i;
		}
    }
	else
	{
		return 0;
	}
}


/****************************************************************************************
* Function name: CircularBufferQuerySpace
* Purpose: get the free space in the circular buffer
*****************************************************************************************/
u4 CircularBufferQuerySpace(u4 bufferId)
{
	return  CbufSize[bufferId] - (CbufWritePos[bufferId] + (CbufReadPos[bufferId] > CbufWritePos[bufferId]) * CbufSize[bufferId] - CbufReadPos[bufferId]);
}

/****************************************************************************************
* Function name: CircularBufferQuerySpace
* Purpose: get the used space in the circular buffer
*****************************************************************************************/
u4 CircularBufferQueryUsed(u4 bufferId)
{
	return CbufSize[bufferId] - CircularBufferQuerySpace(bufferId);
}

/****************************************************************************************
* Function name: CircularBufferRemoveLastWrittenValues
* Purpose: remove the last written values from the circular buffer
*****************************************************************************************/
void CircularBufferRemoveLastWrittenValues(u4 bufferId, u4 size)
{
	if(size > CbufWritePos[bufferId])
	{
		CbufWritePos[bufferId] = CbufSize[bufferId] - (size - CbufWritePos[bufferId]);
	}
	else
	{
		CbufWritePos[bufferId] -= size;
	}
}

/****************************************************************************************
* Function name: CircularBufferRemove
* Purpose: remove values from the circular buffer (this is equivalent to a read but
there is no copy of buffers)
*****************************************************************************************/
void CircularBufferRemove(u4 bufferId, u4 size)
{
	if (CircularBufferQueryUsed(bufferId) >= size)
    {
		if ((CbufSize[bufferId] - CbufReadPos[bufferId]) >= size) 
		{
			CbufReadPos[bufferId] += size;
		}
		else
		{
			CbufReadPos[bufferId] += size - CbufSize[bufferId]; 
		}
    }
}
