/****************************************************************************************
* File name: ramp_ft.c
* Purpose:  Ramp Fitting.		 
*
* Description:	The Ramp Fitting is used in Spectroscopy Mode to 
*				Calculate Ramps Slope.
****************************************************************************************/

#ifdef SIMULATOR
#include "virtuosoSim.h"
#else // SIMULATOR
#include "node1.h"
#include "k_struct.h"
#endif
#include "spuAsw.h"
#include "spulib.h"

/*=================================================================================================*/
/*Functions declared in this file
===================================================================================================*/

void RampFit(u4 readoutsPerRamps, int PixelsForRounding, int FramesForRounding, u4 CompressionMode);
void LeastSq_With_Glitch_Det(u4 *Address, u4 readoutsPerRamps); 
void LeastSquareWithoutGlitchDetection(u4 *Address, u4 readoutsPerRamps); 
void TwoSamples(u4 *buf);
void MeanvalueWithoutGlitchDetection(u4 *Addr, u4 readoutsPerRamps);
void Meanvalue_with_glitch_det(u4 *Addr, u4 readoutsPerRamps);

/*=================================================================================================*/
/*External function and variables
===================================================================================================*/
extern int DecimatePixelOrder (int *buffer, u4 framesize, u4 frames, u4 groupsize, u4 discard_pre, u4 discard_post);
extern int StatAveraging (int *source, int nbPixelsPerFrame, int nbFrames, int *dest, int numavg, int RoundingBits, int strategy);
void ClampSpec (int *source, int nsamples, int RoundingBits);

extern int ReductionStatus;
extern float SumOfX[512], LsqDenominator[512]; /* Look up Tables for the Least Square Fit */
/*=================================================================================================*/


/*
This function performs the ramp fitting using one of the many algorithms available
*/
void RampFit(u4 readoutsPerRamps, int PixelsForRounding, int FramesForRounding, u4 CompressionMode)
{
	u4 bufferSize = ProcBuff[0];
	u4 newBufferSize;
	
	/*=============================================================================================*/
	/* Initialization of the HK Parameters of Compression Information
	===============================================================================================*/
	
	Result.NbMaintainedRamps = 0;
	Result.SaturationFlag = 0;
	Result.GlitchCounter = 0;
	Result.NbIntegratedRamps = 0;
	
	/*=============================================================================================*/
	/* The Least Square Fitting is Performed after a Glitch Removal*/
	
	/* Fitting over DetectorConstantsSpectro.ReadoutsPerSubRamps Samples */
	if(DetectorConstantsSpectro.ReadoutsPerSubRamps > 512)
	{
		DetectorConstantsSpectro.ReadoutsPerSubRamps = 32;
	}
	if(DetectorConstantsSpectro.ReadoutsPerSubRamps != 0)
	{
		newBufferSize = bufferSize/DetectorConstantsSpectro.ReadoutsPerSubRamps; /* The New Size after the Fitting is Calculated */
	}
	else
	{
		newBufferSize = bufferSize >> 5;
	}
	
	/* Write the New Data Size on the Top Address After Ramp Fitting */
	ProcBuff[0] = newBufferSize;   
	
	/* We don't check for Signal Saturation anymore, since the DMC ADC has an higher range than the CRE output voltage, it can never saturate the ADC */
	Result.SaturationFlag = 0;
	
	ReductionStatus=0; // for hk
	
	/* Mean Algorithm*/
	if(DetectorConstantsSpectro.RampFittingAlgorithm == K_MEAN_ALGO)
	{
		/* Call the Mean Function*/
		if(DetectorConstantsSpectro.UseGlitchDetection == K_YES)
			/* With Glitch Detection*/
		{
			Meanvalue_with_glitch_det(ProcBuff, DetectorConstantsSpectro.ReadoutsPerSubRamps);
			Result.ReductionAlgo = 0x12;
		}
		else /* Without Glitch Detection*/
		{
			if (DetectorConstantsSpectro.RoundingStrategy == K_ROUNDING_STYLE_138)
			{
				MeanvalueWithoutGlitchDetection(ProcBuff, DetectorConstantsSpectro.ReadoutsPerSubRamps);
    
			}
			else // new 13.9 averaging with rounding and 13.10 decimation
			{
			 	/* new reduction step in version 13.10: frame dropping */
				FramesForRounding = DecimatePixelOrder ((int*)ProcBuff+1, PixelsForRounding, FramesForRounding, readoutsPerRamps, DetectorConstantsSpectro.DiscardPre, DetectorConstantsSpectro.DiscardPost);
				ProcBuff[0] = FramesForRounding * PixelsForRounding; // update buffer size

				if (DetectorConstantsSpectro.RoundingStrategy == K_ROUNDING_STYLE_ROLAND)
				{
					ProcBuff[0] = StatAveraging (ProcBuff+1, PixelsForRounding, FramesForRounding, ProcBuff+1, DetectorConstantsSpectro.ReadoutsPerSubRamps, DetectorConstantsSpectro.RoundingBits, K_ROUNDING_STYLE_ROLAND);
				}
				else // koryo style
				{
					ProcBuff[0] = StatAveraging (ProcBuff+1, PixelsForRounding, FramesForRounding, ProcBuff+1, DetectorConstantsSpectro.ReadoutsPerSubRamps, DetectorConstantsSpectro.RoundingBits, K_ROUNDING_STYLE_KORYO);
				}
				ClampSpec ((int*)ProcBuff+1, ProcBuff[0], DetectorConstantsSpectro.RoundingBits);

				// note: the shift bit info is downlinked in write.c (PreProcessingParametersSummary...) and not here
				Result.NbMaintainedRamps = roldiv(ProcBuff[0], PixelsForRounding);
			}
			Result.ReductionAlgo = 0x13;
			ReductionStatus = DetectorConstantsSpectro.RoundingBits;
		}
	}	
	else	//this is the least square algo
	{
		if(DetectorConstantsSpectro.ReadoutsPerSubRamps == 2)
			/* Case of 2 Samples per Sub-Ramp Then Use Sample Difference*/
		{
			TwoSamples(ProcBuff);
			Result.ReductionAlgo = 0x14;
		}
		else
		{
			/* Call the Least Square Fit Function*/
			if (DetectorConstantsSpectro.UseGlitchDetection == K_YES)
				/* With Glitch Detection*/
			{
				LeastSq_With_Glitch_Det(ProcBuff, DetectorConstantsSpectro.ReadoutsPerSubRamps);
				Result.ReductionAlgo = 0x01;
			}
			else
				/* Without Glitch Detection*/			
			{
				LeastSquareWithoutGlitchDetection(ProcBuff, DetectorConstantsSpectro.ReadoutsPerSubRamps);
				Result.ReductionAlgo = 0x11;
			}
		}
	}
	
	/* HK Parameter Defining the Reduction Algorithm Used */
	/* for 13.10 the hi-byte is used for decimation and sync delay instead of the subramp length */
	/* the subramp length is anyway transported via the NULLWORD (compressedEntityBuffer[2]) */
	/* the HK will overwrite this hi-byte with its own info */
	//Result.ReductionAlgo = Result.ReductionAlgo | (DetectorConstantsSpectro.ReadoutsPerSubRamps<<8);
	Result.ReductionAlgo |= (DetectorConstantsSpectro.DiscardPre & 0x0f) << 12;
	Result.ReductionAlgo |= (DetectorConstantsSpectro.DiscardPost & 0x0f) << 8;
}

/*****************************************************************************
* Function name: void LeastSq_With_Glitch_Det 
*
* Purpose: Calculate the Slope Value of N Samples
*
* Description:	This Function used for the Slope Fitting of Ramp using
*               the Linear Regression Method. It uses Glitch Detection for 
*				Outliers Removal. 
*
* Syntax: void LeastSq_With_Glitch_Det(u4 *Addr, u4 RRR)
*
* Input: Addr: start address, and Number of Samples Per Ramp
*
* Output: Slopes are Written in the Original Buffer
*
* Return: None
*
* Files Accessed: None
*
* Creation date & author: 21.11.2001 by Ahmed Nabil Belbachir
*
* Version, Update date & author:HPSPU-HLSW-OB-V7R0 01.12.2003 by Ahmed Nabil Belbachir
*															  &	 Roland Ottensamer		
*														
********************************************************************************/

void LeastSq_With_Glitch_Det(u4 *buffer, u4 readoutsPerRamps) 
{  
	int threshold = DetectorConstantsSpectro.ThresholdForSpectro;
	
	u4 i,l,n,n0,k=0; 
	u4 KK=0, ramps,new_index=0;
	int j,limit,*Addr,Level,ll=0,Valuta,new_value,New_RRR=readoutsPerRamps,*y; 
	float *sumX, sumY, sumXY, *denominator; /* Intermediate Values */
	float slope_z, slope_n, result=0; /* Intermediate Results */
	
	/*=============================================================================================*/
	
	/*Initialization of the Pointers */
	sumX = SumOfX;
	denominator = LsqDenominator;
	/* Size of the Data*/
	ramps =buffer[0];
	
	/*Pointer to the Data*/
	y=(int*)(buffer+1);
	
	/* A Ramp is split into fields of 4 Samples each */
	limit = readoutsPerRamps>>2;
	
	/* Glitch Rejection and Slope Calculation for Every Ramp/Sub-ramp */
	for(i=0; i<ramps; i++) 
	{
		n=i*readoutsPerRamps;
		Addr = (int *)(y+n);
		
		/* "3" is the index difference bewteen samples at the begining of a Ramp*/
		Valuta=(y[n+1]-y[n])*3;
		if(Valuta>=0)
		{
			Level= 1;
		}
		else
		{
			Level= -1;
		}
		
		sumY = sumXY = 0;  
		new_value=y[n+3]-y[n]; 
		
		/* Glitch Detection and Rejection for each Ramp */
		for (j=0; j<limit; j++)
		{
			n0 =n+(4*j)+2;
			ll= (new_value-Valuta)*Level;
			
			/* "5" is the index difference bewteen samples at the begining of a Ramp*/
			Valuta=(y[n0+1]-y[n0])*5;
			new_value=y[n0+5]-y[n0]; 
			if( ll> threshold) /* Test the Slope and the Ramp Linearity*/
			{
				/* Fit the Safe Samples if Glitch detected */
				/* Case the Glitch appears in the first quadruple */
				if(j==0)
				{
					new_index=4;
				}
				else
				{
					/* Calculate the Number of Samples to Fit before the glitch*/
					KK=(4*j)-KK-new_index;
				}
				
				k++;
				
				/* Intermediate Sums */
				for (l=0; l<KK; l++) 
				{
					sumXY += l*Addr[l]; 
					sumY += Addr[l]; 
				} 
				/* Intermediate Results */
				slope_z = (KK * sumXY - sumY * sumX[KK-1]); 
				slope_n = denominator[KK-1]; 
				
				if(slope_n!=0)
				{
					result += (slope_z/slope_n)*readoutsPerRamps;
				}
				
				/* Ignore the 8 Frames Following a Glich */
				new_index += 8;
				Addr = (int*)(y+n0+1+new_index);				
				New_RRR = New_RRR-new_index-KK;
				if(New_RRR<0)
				{
					New_RRR=0;
				}
				
				/* ignore the next 8 samples if glitch */
				j+=2;
				
				/* 3 is the index difference bewteen samples at the begining of a Ramp*/
				Valuta=(y[n0+10]-y[n0+9])*3;
				new_value=y[n0+12]-y[n0+9]; 
				sumY = sumXY = 0;  
				
				/* Increment the Glitch Counter*/
				Result.GlitchCounter++;
			}
		}
		/* Intermediate Sums */
		for (j=0; j<New_RRR; j++) 
		{
			sumXY += j*Addr[j]; 
			sumY += Addr[j]; 
		}
		/* Intermediate Results */
		slope_z = (New_RRR * sumXY - sumY * sumX[New_RRR -1]); 
		slope_n = denominator[New_RRR-1]; 
		
		if(slope_z>0) /* Choose the Rounding Error for the Division */
		{
			/* Write The Final Result into Buffer */
			buffer[i+1] = (int)((slope_z/slope_n)*readoutsPerRamps+0.5)&0xffff;
		}
		else
		{
			/* Write The Final Result into Buffer */
			buffer[i+1] = (int)((slope_z/slope_n)*readoutsPerRamps-0.5)&0xffff;
		}
		
		if(k!=0)
		{
			/* Average the ntermediate Results for Sub-Ramps if Glitch Occured */
			buffer[i+1] = roldiv(((int)(result)&0xffff)+buffer[i+1], (k+1));
			k=0;
			result=0;
			KK=0;
			New_RRR=readoutsPerRamps;
			
			/* This is to Show Where to Begin the Fit. It is 0 in Case of No Glitch in Ramp */
			new_index=0;	  	 
		}
	}
	/* The Saturation Flag is the Most Significant Byte set to 0x11 is Saturation and to 0 Otherwise*/
	Result.GlitchCounter &=0xFFFFFF;
	
	/* Provide with the Number of Maintained Ramps*/
	Result.NbMaintainedRamps=i;
}

/*****************************************************************************
* Function name: void LeastSquareWithoutGlitchDetection 
*
* Purpose: Calculate the Slope Value of N Samples
*
* Description:	This Function used for the Slope Fitting of Ramp using
*               the Linear Regression Method. It does not use Glitch Detection
*
* Syntax: void LeastSquareWithoutGlitchDetection(u4 *Addr, u4 RRR)
*
* Input: Addr: start address, and Number of Samples Per Ramp
*
* Output: Slopes are Written in the Original Buffer
*
* Return: None
*
* Files Accessed: None
*
* Creation date & author: 21.11.2003 by Ahmed Nabil Belbachir
*
* Version, Update date & author:HPSPU-HLSW-OB-V7R0 01.12.2003 by Ahmed Nabil Belbachir
*															  
********************************************************************************/
void LeastSquareWithoutGlitchDetection(u4 *buffer, u4 nbReadoutsPerSubRamps) 
{
	u4 i,j, nbRamps;
	int *y; 
	float slope_z; /* Intermediate Results */
	
	float floatNbReadoutsPerSubRamps = (float)nbReadoutsPerSubRamps;
	float floatInvertedLsqDenominator = 1.0f/LsqDenominator[nbReadoutsPerSubRamps-1];
	float floatSumOfX = SumOfX[nbReadoutsPerSubRamps -1];
	int sumY, sumXY;
	float floatSumY, floatSumXY; 
	
	/*=============================================================================================*/
	
	/* Size of the Data*/
	nbRamps =buffer[0];
	
	/*Pointer to the Data*/
	y=(int*)(buffer+1);
	
	
	if (nbReadoutsPerSubRamps < 256)
	{
		/* Slope Calculation for Every Ramp/Sub-ramp */
		for(i=0; i < nbRamps; i++) 
		{
			sumY = sumXY = 0;  
			
			/* Intermediate Sums (we work with integers here)*/
			for (j=0; j<nbReadoutsPerSubRamps; j++) 
			{
				sumXY += j*y[j]; 
				sumY += y[j]; 
			}
			
			//convert the integer results to float
			floatSumXY = (float)sumXY;
			floatSumY = (float)sumY;
			
			/* Intermediate Results */
			slope_z = (floatNbReadoutsPerSubRamps * floatSumXY - floatSumY * floatSumOfX); 
			
			if(slope_z>0) /* Choose the Rounding Error for the Division */
			{
				/* Write The Final Result into Buffer */
				buffer[i+1] = (int)(slope_z*floatInvertedLsqDenominator*floatNbReadoutsPerSubRamps+0.5) & 0xffff;
			}
			else
			{
				/* Write The Final Result into Buffer */
				buffer[i+1] = (int)(slope_z*floatInvertedLsqDenominator*floatNbReadoutsPerSubRamps-0.5) & 0xffff;
			}
			
			
			y += nbReadoutsPerSubRamps;
		}
	}
	else	//in the case of big nbReadoutsPerSubRamps, SumXY could overflow.  In this case, we can use
	{		//the old code that works with float.  It is more time consuming but the losless compression
		//will go faster (since the data will be reduced).
		
		float sumY, sumXY; /* Intermediate Values */
		float slope_z, slope_n; /* Intermediate Results */
		
		/* Slope Calculation for Every Ramp/Sub-ramp */
		for(i=0; i < nbRamps; i++) 
		{
			sumY = sumXY = 0;  
			
			/* Intermediate Sums */
			for (j=0; j<nbReadoutsPerSubRamps; j++) 
			{
				sumXY += j*y[j]; 
				sumY += y[j]; 
			}
			/* Intermediate Results */
			slope_z = (nbReadoutsPerSubRamps * sumXY - sumY * SumOfX[nbReadoutsPerSubRamps -1]); 
			slope_n = LsqDenominator[nbReadoutsPerSubRamps-1]; 
			
			if(slope_z>0) /* Choose the Rounding Error for the Division */
			{
				/* Write The Final Result into Buffer */
				buffer[i+1] = (int)((slope_z/slope_n)*nbReadoutsPerSubRamps+0.5)&0xffff;
			}
			else
			{
				/* Write The Final Result into Buffer */
				buffer[i+1] = (int)((slope_z/slope_n)*nbReadoutsPerSubRamps-0.5)&0xffff;
			}
			y += nbReadoutsPerSubRamps;
		}
	}
	
	/* The Saturation Flag is the Most Significant Byte set to 0x11 is Saturation and to 0 Otherwise*/
	Result.GlitchCounter &=0xFFFFFF;
	
	/* Provide with the Number of Maintained Ramps*/
	Result.NbMaintainedRamps=i;
}

/*****************************************************************************
* Function name: TwoSamples
*
* Purpose: Calculate the Slope Value of 2 Samples
*
* Description:	This Function used for the Slope Fitting by calculating the 
*				difference between Two Adjacent samples 
*
* Syntax: void TwoSamples(u4 *buf)
*
* Input: Addr: start address
*
* Return: Result are Returned in the Original Buffer
*
* Files Accessed: None
*
* Creation date & author: 21.11.2001 by Ahmed Nabil Belbachir
*
* Version, Update date & author:HPSPU-HLSW-OB-V1R0 16.09.2002 by Ahmed Nabil Belbachir
*														
********************************************************************************/

void TwoSamples(u4 *buf)
{
	u4 i,ii;
	for (i=1;i<=buf[0];i++) /* Calculate the Difference between 2 Successive Samples*/
	{	
		ii=i*2;
		buf[i] = buf[ii-1] - buf[ii]; // Send the Positive Value of the Slope
	}
	
	/* Write the Results to HK*/
	Result.GlitchCounter &=0xFFFFFF;
	Result.SaturationFlag=0;
	
	/* Provide with the Number of Maintained Ramps*/
	Result.NbMaintainedRamps = buf[0];
}

/*****************************************************************************
* Function name: MeanvalueWithoutGlitchDetection
*
* Purpose: Calculate the Mean Value of N Samples
*
* Description:	This Function used in the Robust Averaging in Photometry.
*               It Calculates the Mean Value of N Samples 
*
* Syntax: void MeanvalueWithoutGlitchDetection(u4 *Address, u4 Number_of_Samples_To Average)
*
* Input: start address
*
* Output: Number of Sample to Average
*
* Return: Result are Returned in the Original Buffer
*
* Files Accessed: None
*
* Creation date & author: 21.11.2001 by Ahmed Nabil Belbachir
*
* Version, Update date & author:HPSPU-HLSW-OB-V1R0 20.02.2004 by Ahmed Nabil Belbachir
*														
********************************************************************************/

void MeanvalueWithoutGlitchDetection(u4 *buf, u4 readoutsPerSubramps)
{
	u4 nbShifts = DetectorConstantsSpectro.Log2OfReadoutsPerSubRamps;
	u4 i, sum, firstPixelOfSubramp, *y, j;
	u4 nbSubramps = buf[0];
	y = buf+1;
	
	if(readoutsPerSubramps != 0)
	{		
		for (i=0; i < nbSubramps; i++)		  /* Averaging each RRR Samples */
		{
			sum = 0;
			firstPixelOfSubramp = readoutsPerSubramps * i;
			for(j=0; j < readoutsPerSubramps; j++) 
			{
				sum += y[firstPixelOfSubramp + j];
			}
			y[i] = (sum + (u4)RoundingHelper[nbShifts]) >> nbShifts; /* Average Over each RRR Samples */
		}
	}	
	/* Write the Results to HK*/
	Result.GlitchCounter &= 0xFFFFFF;
	
	/* Provide with the Number of Maintained Ramps*/
	Result.NbMaintainedRamps = nbSubramps;
	
}

/*****************************************************************************
* Function name: Meanvalue_with_glitch_det
*
* Purpose: Calculate the Mean Value of N Samples
*
* Description:	This Function used in the Robust Averaging in Photometry.
*               It Calculates the Mean Value of N Samples 
*
* Syntax: void Meanvalue_with_glitch_det(u4 *Address, u4 Number_of_Samples_To Average)
*
* Input: start address
*
* Output: Number of Sample to Average
*
* Return: Result are Returned in the Original Buffer
*
* Files Accessed: None
*
* Creation date & author: 21.11.2001 by Ahmed Nabil Belbachir
*
* Version, Update date & author:HPSPU-HLSW-OB-V11R0 17.07.2004 by Ahmed Nabil Belbachir
*														
********************************************************************************/

/* Begin of the Meanvalue_with_glitch_det Function 
===================================================================================================*/
void Meanvalue_with_glitch_det(u4 *Addr, u4 readoutsPerRamps)
{
	u4 i, Summ,j,factor,sill;
	
	int threshold = DetectorConstantsSpectro.ThresholdForSpectro;
	int diff;
	
	if(readoutsPerRamps!=0)
	{
		for (i=0;i<Addr[0];i++)		  /* Averaging each RRR Non-Glitched Samples after Glitch Removal */
		{
			factor=0;
			sill= (readoutsPerRamps*i)+1;
			/* Check the Glitch in the First two Samples */
			diff= Addr[sill]-Addr[sill+1];
			if(diff < threshold)
			{	
				Summ=Addr[sill]+Addr[sill+1];
				factor=2;
			}
			else
			{
				/* Discard the First two Samples if Glitch*/
				Summ=0;
				
				/* Increment the Glitch Counter*/
				Result.GlitchCounter +=2;
			}
			
			/* Check the Glitch in the Intermediate Samples */
			for(j=2;j<readoutsPerRamps;j++) 
			{
				diff= Addr[sill+j-1]-Addr[sill+j];
				if(diff < threshold)
				{	
					Summ += Addr[sill+j];
					factor++;
				}
				/* Discard every two Successive Samples if Glitch*/
				else
				{	j++;
				/* Increment the Glitch Counter*/
				Result.GlitchCounter++;
				}
			}
			
			/* If all Ramp Samples all Glitched*/
			if (factor==0)
			{
				Addr[i+1] = 0; 
			}
			else
			{
				Addr[i+1] = Summ/factor; /* Average Over "factor" Samples */
			}
		}
		
		/* Write the Results to HK*/
		Result.GlitchCounter &=0xFFFFFF;
		
		/* Provide with the Number of Maintained Ramps*/
		Result.NbMaintainedRamps=i;
	}
}
