/****************************************************************************************
* File name: spulib.h
* Description:	This Header File Contains the Memory Map, the DPU Interfacing
*				Parameters and the Data Packet Structures.
****************************************************************************************/

#ifndef _spulib_H
#define _spulib_H

#include "genspu.h"	

/*===========================================================================================================*/
/* Declaration of the Acknowledgment Codes
=============================================================================================================*/
 
typedef enum
{

	K_PACK_LOAD				= 0x00810000,	
	K_PACK_DUMP_INTER		= 0x01820000,    /* Pack for the First and  Intermediate Block*/
	K_PACK_DUMP_LAST		= 0x00820000,	  /* Pack for the Last Block*/
	K_PACK_CHECK			= 0x00830000,
	K_PACK_PERFORM			= 0x00840000,
	K_PACK_WRITE			= 0x00860000,

	K_NACK_LOAD				= 0x00F10000,	
	K_NACK_DUMP				= 0x00F20000,
	K_NACK_CHECK			= 0x00F30000,
	K_NACK_PERFORM			= 0x00F40000,
	K_NACK_WRITE			= 0x00F60000,
	K_NACK_INVALID_CMD		= 0x01FF0000,

} SPU_ACKS;


/*===========================================================================================================*/
/* Declaration of the Nack Error Codes 
=============================================================================================================*/

typedef enum
{
	K_INVALID_CMD_ID		= 0x0071,
	K_INVALID_MEM_ID		= 0x0072,
	K_INVALID_ADDRESS		= 0x0073,	
	K_INVALID_LENGTH		= 0x0074,
	K_INVALID_CHECKSUM		= 0x0075,
	K_INVALID_CHECKSUM_1	= 0x0075,
	K_INVALID_CHECKSUM_2	= 0x007B,
	K_INVALID_PARAMETER_ID	= 0x0076,
	K_INVALID_ACTIVITY_ID	= 0x0077,
	K_INVALID_SID			= 0x0078,
	K_INVALID_SID_PARAMETER	= 0x0079,

} SPU_NACK_ERROR;

/*===========================================================================================================*/
/* Useful Memory Addresses 
=============================================================================================================*/

typedef enum
{

	K_DM_END_DRAM				= 0x07FFFF,
	K_DM_END_DPRAM				= 0x01FFF,
	K_PM_START_EEPROM			= 0xE0000,
	K_PM_END_EEPROM				= 0x11FFFF,
	K_MAX_LOAD_LENGTH			= 51,
	K_MAX_DUMP_LENGTH_IN_DM		= 249,
	K_MAX_DUMP_LENGTH_IN_PM		= 166,
	K_WRITE_CMD_LENGTH			= 0x0018,

	K_DX_BLUE_PHOT_LENGTH		= 2049,
	K_DX_RED_PHOT_LENGTH		= 513,
	K_DX_SPECTRO_LENGTH			= 469,
	K_DX_TABLE_MAX_SIZE			= 2048,

} MEMORY_MAP;

/*===========================================================================================================*/
/* Compression modes 
=============================================================================================================*/

typedef enum
{
	K_CM_UNDEFINED							= 0xFF,

	K_CM_PHOTO_DEFAULT						= 0x00,
	K_CM_PHOTO_DOUBLE_COMPRESSION			= 0x01,
	K_CM_PHOTO_HALF_COMPRESSION				= 0x02,
	K_CM_PHOTO_LOSSLESS_COMPRESSION			= 0x04,	
	K_CM_PHOTO_TRANSPARENT					= 0x07, 
	K_CM_PHOTO_BUFFER_TRANSMISSION			= 0x09,
	K_CM_PHOTO_COMPRESSED_SENSING			= 0x0C,

	K_CM_SPECTRO_DEFAULT					= 0x10,
	K_CM_SPECTRO_LOSSLESS_COMPRESSION		= 0x14,
	K_CM_SPECTRO_TRANSPARENT				= 0x17,
	K_CM_SPECTRO_4_SEC_RESET				= 0x18,
	K_CM_SPECTRO_BUFFER_TRANSMISSION		= 0x19,


} COMPRESSION_MODES;


/*===========================================================================================================*/
/* Circular buffers identifiers
=============================================================================================================*/

typedef enum
{
	K_CBUF_TM,
	K_CBUF_HEADER,
	K_CBUF_SCIENCE_BLUE_PHOT1,
	K_CBUF_SCIENCE_BLUE_PHOT2,
	K_CBUF_SCIENCE_BLUE_PHOT3,
	K_CBUF_SCIENCE_BLUE_PHOT4,
	K_CBUF_SCIENCE_RED_PHOT,
	K_CBUF_SCIENCE_SPECTRO = K_CBUF_SCIENCE_RED_PHOT,

	K_CBUF_MEM_SCRUB_PM_SF,	//memory scrubbing results PRAM, single bit failures
	K_CBUF_MEM_SCRUB_PM_DF,	//memory scrubbing results PRAM, double bit failures
	K_CBUF_MEM_SCRUB_DM_SF,	//memory scrubbing results DRAM, single bit failures
	K_CBUF_MEM_SCRUB_DM_DF,	//memory scrubbing results DRAM, double bit failures

	K_CBUF_DMC_HEADER_ERRORS,
	K_CBUF_NB_CIRCULAR_BUFFERS
} CIRCULAR_BUFFERS_ID;

/*===========================================================================================================*/
/* Declaration of the HK Packet
=============================================================================================================*/

struct SPU_HK_PACKET			/* The parameters are defined in 4 bytes word */
{
	u4 Header;	                /* HK Header */
	u4 ObsId;					/* OBS ID */
	u4 PacketIndex;				/* Packet Index */
	u4 CounterInc;				/* Counter Increment */
	u4 ReductionAlgo;			/* Reduction Algorithm */
	u4 SaturationFlag;			/* Saturation Flag */
	u4 GlitchCounter;			/* Glitch Counter Information */
	u4 NbMaintainedRamps;		/* Number of Maintained Ramps/Averages */
	u4 CpuWorkload;				/* CPU WorkLoad*/
	u4 DmcLinkStatus;			/* Status of the DEC/MEC Link Connection: 0xFF=off, 0xAA=in_progress, 0=on */
	u4 NbIntegratedRamps;		/* Number of Integrated Ramps */
	u4 VersionId;				/* Version ID of the SW */
	u4 RawChannelIndex;			/* Raw Channel Index */
	u4 DmcHeaderError;			/* DMC Header Error */
  	u4 MemErrorCounter;			/* Parameters Counting the Corrupted Memory Addresses */
	u4 Spare;					/* Spare */
	u4 LlcError;				/* Lossless Compression Error */
	u4 ParameterWished;			/* Address of wished parameter to visualize */
};


/*===========================================================================================================*/
/* Declaration of the Packet to Transmit
=============================================================================================================*/

struct FIFO_TRANSMIT_MSG
{
	u4	Header;					/* Response Packet Header */
	union
	{	
		u4 Size;				/* Data Size*/
		u4 ObsId;				/* OBS ID */
		u4 Counter;				/* Blocks Counter */
		u4 AddressLength;		/* Address1 in 2-MSB Bytes and Length in 2-LSB Bytes */
		u4 NackParameter;		/* Parameter for the Negative Acknowledgment */
	}Block1;

	u4 Checksum;			/* CheckSum in 2-MSB Bytes*/
	u4	NbPacketsToTransmit;	/* for compressed science data, this contains the remaining number of packets to be transmitted */
} ;



/*===========================================================================================================*/
/* Structue of the parameter of the WRT_SIM_DATA write command
=============================================================================================================*/
typedef struct WRT_SIM_DATA
{
	u4	SpuId;
	u4	CompressionMode;
	u4	ReadoutsPerRamps;
	u4	Spares1[15];
	u4	DmcHeaderCompressionAlgorithm;	// 0 = rzip2 (new), 1 = rzip32 (old)
	u4	ComputeChecksumOnScienceData;	
	u4	ProgrammableHkAddress;			//user can write an address that will be monitored in the hk (SPU_PAR_MONITOR)
	u4	Spares2[3];
} WRT_SIM_DATA;

//defines used in WRT_DET_CST_Type
#define K_ROUNDING_STYLE_ROLAND		0x00
#define K_ROUNDING_STYLE_KORYO		0x01
#define K_ROUNDING_STYLE_138		0x02

/*===========================================================================================================*/
/* Structue of the parameter of the WRT_DET_CST_SPECTRO write command
=============================================================================================================*/
typedef struct WRT_DET_CST_SPECTRO
{
	u4	PreProcessingFlag;				//0= no, 1 = add, 2 = sub, 3 = mult, 4 = div
	u4	PreProcessingParameter;
	int	ThresholdForSpectro;			//this param is used for glitch detection (it defines the ramp linearity)
	int	ThresholdForChopper;
	u4	ThresholdForGrating;
	u4	ReadoutsPerSubRamps;
	u4	UseGlitchDetection;				//0 = use; 1 = don't use
	u4	RampFittingAlgorithm;			//0 = least square; 1 = mean
	u4	SortingAlgorithm;			
	u4	LosslessCompressionAlgorithm;
	u4	PacsCodecOrder;
	u4  RoundingBits;
	u4  RoundingStrategy;
	u4  FixRawChannels;
	u4  DiscardPre;
	u4  DiscardPost;
	u4	Spares[8];

	// these parameters can not be affected by a write command
	u4	Log2OfReadoutsPerSubRamps;
} WRT_DET_CST_SPECTRO;

/*===========================================================================================================*/
/* Structue of the parameter of the WRT_DET_CST_PHOTO write command
=============================================================================================================*/
typedef struct WRT_DET_CST_PHOTO
{
	u4	PreProcessingFlag;				//0= no, 1 = add, 2 = sub, 3 = mult, 4 = div
	u4	PreProcessingParameter;
	int	ThresholdForPhoto;				//this param is used for glitch detection
	int	ThresholdForChopper;
	u4	UseGlitchDetection;				//0 = use; 1 = don't use
	u4	NbSamplesToAverage;
	u4	RoundingBits;
	u4	RoundingStrategy;
	u4 	Scheme;							// 1 = velvet; 0 = velveteen
	u4	LosslessCompressionAlgorithm;
	u4	PacsCodecOrder;
	u4  DiscardPre;
	u4  DiscardPost;
	u4 	SyncDelay;
	u4	Spares2[10];
} WRT_DET_CST_PHOTO;


/*===========================================================================================================*/
/* Structue representing a detector selection table
=============================================================================================================*/
typedef struct DXS_TABLE
{
	u4	NbSelectedPixels;
	u4	Table[K_DX_TABLE_MAX_SIZE];
	u4	Size; 
	u4	NbSelectedPixelsInEachPhotoBlock[4]; //these variables are used only for the Blue Photometer DXS
} DXS_TABLE;

/*===========================================================================================================*/
/* Structue containing information about the raw channels selection
=============================================================================================================*/
typedef struct RAW_CHANNEL_SELECTION
{
	u4	NbSelectedRawChannels;
	u4	IndexOfFirstChannel;
} RAW_CHANNEL_SELECTION;

/*===========================================================================================================*/
/* Structue containing the raw channels data (before and after compression)
=============================================================================================================*/
typedef struct RAW_CHANNEL_DATA
{
	u4	SizeInWords;
	u4	Data[0x1500];
	u4	CompressedData[0x1500];
} RAW_CHANNEL_DATA;

/*===========================================================================================================*/
/* Structue containing the compressed DMC headers
=============================================================================================================*/
typedef struct COMPRESSED_DMC_HEADER_BUFFER
{
	u4	SizeInWords;
	u4	Data[0x1500];
} COMPRESSED_DMC_HEADER_BUFFER;

/*===========================================================================================================*/
/* Structue used to describe a buffer
=============================================================================================================*/
typedef struct BUFFER_DESCRIPTION
{
	u4*	Data;
	u4	SizeInWords;
	u4	Type;
} BUFFER_DESCRIPTION;

//defines used in BUFFER_DESCRIPTION.Type
#define K_DATA_TYPE_2_SHORTS_PER_WORD				0x00
#define K_DATA_TYPE_1_SHORT_PER_WORD				0x01
#define K_DATA_TYPE_1_SHORT_PER_WORD_MAPPED_TO_POS	0x03

/*===========================================================================================================*/
/* Structue containing all the information about the memory scrubbing
   Note: this structure is used by Assembly code, if you reorganize it, you should also change the offset
   in smcs_isr.s
=============================================================================================================*/
typedef struct MEMORY_SCRUBBING_PARAMS
{
	u4	CurrentDmAddress;
	u4	CurrentPmAddress;

	u4 indexNextSfDmCorruptedAddress;
	u4 indexNextSfPmCorruptedAddress;
	u4 indexNextDfDmCorruptedAddress;
	u4 indexNextDfPmCorruptedAddress;

	u4	SingleFailureCounter;
	u4	DoubleFailureCounter;
} MEMORY_SCRUBBING_PARAMS;

#include "globals.h"

/*===========================================================================================================*/
#endif