/****************************************************************************************
* File name: supervs.c
* Purpose: Records the data and triggers the compression.
*
* Description:	The Supervisor is the Controller of the Application Software.
*               As soon as the compression software has been started, it is recording the
*               data from DMC.
*               As soon as a buffer is full or when a transition occurs, it triggers the compression
*               of the buffer and start recording the next one.
****************************************************************************************/

#ifdef SIMULATOR
#include "virtuosoSim.h"
#else // SIMULATOR
#include "node1.h"
#include "k_struct.h"
#endif
#include "spuAsw.h"
#include "spulib.h"

/*=================================================================================================*/
/*Functions declared in this file
===================================================================================================*/

void Supervisor();
void CompressLastSpectroBuffer();
void CompressLastPhotoBuffer();
void DiscardCurrentPhotoBuffers(u4 spuId);
void DiscardCurrentSpectroBuffers();

/*=================================================================================================*/
/*Global variables declared in this file
===================================================================================================*/

/*Science, DMCH and RC Buffers*/
u4 DmcPacketCounter = 0;						//counts the frame received to know when we need to start compressing a buffer
u4 DmcHeaderStored = 0;							//counts the number of header that have been stored since we started buffering
u4 DmcReadoutsStored[6];						//counts the number of readouts received since we started buffering (this is used when erasing circular buffers).  0 is spectro, 1 to 5 is photo (according to block id)
u4 StoreNow = K_NO;								//this state variable is used to tell the SW to store the science data in a buffer in order to compress it later
u4 BufferTransmissionModeBufferId=0;			//the current buffer being filled

struct SCIENCE_PACKET* SciencePacket;			//the buffer that where the data from DMC is copied
struct FIFO_CSW_MSG CswFifoMsg;					//a message to communicate with the compression task

u4 CurrentCompressionMode = K_CM_UNDEFINED;		//the current compression mode
u4 CurrentDmcLabel = 0;							//the current DMC label
u4 CurrentObsId = 0;							//the current OBSID
u4 CompressedEntityCounter = 0;					//counts the number of compressed entity since the startup of the SPU
u4 CurrentReadoutCountIn4SecondsRamp = 0x200;	//record if we are in the first/second half of the 4sec ramp
u4 Delay = 0;
u4 StartOfSequence = 0;

//these memory pointers are used in the compression algorithms
u4 *mem_1, *mem_2,*tmp_buf;
u4 *mem_3,*mem_4;
u4 *mem_5,*mem_6, *mem_8, *mem_9;

/*=================================================================================================*/
/*External function and variables
===================================================================================================*/

extern struct SCIENCE_PACKET TestSciencePacket;		//used when simulating data on board

#ifndef SIMULATOR
extern void LinkSmcsRead(int device);
extern int ChannelReceiveAddress[3];
#endif 

/*=================================================================================================*/

/*
This is the entry point for the SUPERVISOR task.
*/
void Supervisor()
{
	int i=0;

#ifdef SIMULATOR
	struct SCIENCE_PACKET simulatorSciencePacket;
	SciencePacket = &simulatorSciencePacket;
#else
	SciencePacket = (struct SCIENCE_PACKET*)ChannelReceiveAddress[K_SMCSCHANNEL_2-1];	//SciencePacket points directly in DPRAM
#endif

	CswFifoMsg.Message = K_COMPRESS_DATA;

	/* Reserve Memory for Data Buffering DEC/MEC, Science and TM Data */
    mem_1 = DramBuffer;
	mem_2 = ProcBuff+0x10;
	mem_3 = DramBuffer;
	mem_4 = DramBuffer+0x10000;
	mem_5 = DramBuffer+0x20000;
	mem_6 = DramBuffer+0x20100;
	mem_8 = DramBuffer+0x20200;
	mem_9 = DramBuffer+0x21200;
	tmp_buf = DramBuffer+0x10000;
	ResetCircularBuffers();

	// initialization of buffers
	SortedDataBuffer.Data = ProcBuff+0x21000;

	/*=============================================================================================*/
	
	while (1)
	{		
		/* Listen to the DEC/MEC Link
	 	==========================================================================================*/

#ifdef SIMULATOR
#define TESTING_DATA_GENERATION 0	//set it to 1 to test the data generation on board, set it to 0 otherwise
#if TESTING_DATA_GENERATION
		KS_SemaTestW(19);
		MemCopyDmToDm((u4 *)&TestSciencePacket, (u4*)SciencePacket,272);
#else //TESTING_DATA_GENERATION

		Link1355Read(K_SMCSCHANNEL_2, 272, SciencePacket);	//when testing photo
//		Link1355Read(K_SMCSCHANNEL_2, 250, SciencePacket);	//when testing spectro

#endif //TESTING_DATA_GENERATION
#else //SIMULATOR

		LinkSmcsRead(K_SMCSCHANNEL_2);

#endif //SIMULATOR

		if (CompressionSoftwareRunning==K_YES) // if the compression software is running, look at the data.  Otherwise, ignore them and wait for the next packet
		{
			//perform some basic checks on the DMC header
			if ((SciencePacket->SpuId != 2) && (SciencePacket->SpuId != 3))
			{
				PushDmcHeaderError(K_DMC_INVALID_SPU_ID);
				continue;  //ignore current packet and listen to next packet
			}

			switch(SciencePacket->Type)
			{
				case K_SPECTRO : /* Spectroscopy Mode */
				{  		
					//has the compression mode changed ?
					if (SciencePacket->CompressionMode != CurrentCompressionMode)
					{
						//if the new mode is the buffer transmission mode, we change immediately
						if (CurrentCompressionMode == K_CM_SPECTRO_BUFFER_TRANSMISSION)
						{
							//discard the current buffer (if any)
							DiscardCurrentSpectroBuffers();
							CurrentCompressionMode = SciencePacket->CompressionMode;  //store the current compression mode
						}
						else 
						{	
							//for other mode, we will really change the compression mode only at the beginning of the next ramp
							if((SciencePacket->CurrentReadoutCounter&0x1FF)==((SciencePacket->Block2.ReadoutsPerRamps-1)&0x1FF))
							{
								//if we have already buffered some data, compress them now
								if (DmcPacketCounter > 0)
								{
									//compress and send the last buffer
									CompressLastSpectroBuffer();
									StoreNow = K_NO;
								}
								CurrentCompressionMode = SciencePacket->CompressionMode;  //store the current compression mode
							}
						}
					}

					//if we have not started buffering the data or if we are at the beginning of a new buffer
					if (DmcPacketCounter == 0)
					{
						if (CurrentCompressionMode == K_CM_SPECTRO_BUFFER_TRANSMISSION) /* Case of Buffer Transmission Mode */
						{
							if (BufferTransmissionModeEnabled == K_NO)  //if the SPU is not currently transmitting the data from a previous buffer, we can start a new one
							{
								/* Store the Relevant Information of DEC/MEC Header from 1st Frame */
								CurrentObsId = SciencePacket->ObsId;		/* Get Initial Value of OBS ID */
								StoreNow = K_YES;						//start buffering the data
							}
						}
						else /* Case of Other Compression Modes */
						{
							//start buffering the data only on the first readout of a ramp (note that, for the 4 sec reset interval mode (where ramps have 1024 readouts)
							//we start buffering at the beggining of the ramp and after 512 readouts to have 2 2sec buffers)
							if((SciencePacket->CurrentReadoutCounter&0x1FF)==((SciencePacket->Block2.ReadoutsPerRamps-1)&0x1FF))  
							{
								/* Store the Relevant Information of DEC/MEC Header from 1st Frame */
								CurrentObsId = SciencePacket->ObsId;		/* Get Initial Value of OBS ID */
								StoreNow = K_YES;						//start buffering the data
							}
						}
					}

					//if we are now buffering the data
					if (StoreNow==K_YES)
					{
						//store dmc header and data in circular buffers.  If circular buffers are full, discard the current
						//buffer and wait for the next one
						if (WriteInCircularBuffer(K_CBUF_HEADER, &(SciencePacket->ObsId), K_SPECTRO_HEADER_SIZE) == 0)
						{
							DiscardCurrentSpectroBuffers();
							PushDmcHeaderError(K_DMC_UNABLE_TO_PROCESS_DATA_FAST_ENOUGH);
							continue; //wait for next buffer
						}
						DmcHeaderStored++;

						if (WriteInCircularBuffer(K_CBUF_SCIENCE_SPECTRO, SciencePacket->Data, K_SPECTRO_FRAME_SIZE_IN_WORDS) == 0)
						{
							DiscardCurrentSpectroBuffers(); 
							PushDmcHeaderError(K_DMC_UNABLE_TO_PROCESS_DATA_FAST_ENOUGH);
							continue; //wait for next buffer
						}

						DmcPacketCounter++;		//Increment the Packet Counter
						DmcReadoutsStored[0]++;	//Increment the readout counter

						//Start Compression when a 2 sec buffer is full
						if (DmcPacketCounter == K_NB_SPECTRO_FRAMES_IN_2SEC_BUF)
						{
							if (CurrentCompressionMode == K_CM_SPECTRO_4_SEC_RESET)	//record if we are in the first/second half of the 4sec ramp
							{
								CurrentReadoutCountIn4SecondsRamp = SciencePacket->CurrentReadoutCounter;
							}

							CompressLastSpectroBuffer();

							//stop buffering.  It will restart automatically when everything is in condition again (this might well be at next frame) 
							StoreNow = K_NO;	
						}
					}
				};break;
				case K_PHOTO :  /* Photometry Mode */
				{
					//perform basic checks of DMC header
					if((SciencePacket->Block3.DataBlockId>5) || (SciencePacket->Block3.DataBlockId<1))
					{
						PushDmcHeaderError(K_DMC_INVALID_BLOCK_ID);
						continue;  //ignore current packet and listen to next packet
					}

					//has the compression mode changed ?
					if (SciencePacket->CompressionMode != CurrentCompressionMode)
					{
						//if the new mode is the buffer transmission mode, we change immediately
						if (CurrentCompressionMode == K_CM_PHOTO_BUFFER_TRANSMISSION)
						{
							//discard the current buffer
							DiscardCurrentPhotoBuffers(SciencePacket->SpuId);
							CurrentCompressionMode = SciencePacket->CompressionMode;  //store the current compression mode
						}
						else
						{	
							//for other mode, we will really change the compression mode only at the beginning of the next readout (in case the change occurs while we have received only 1, 2 or 3 packets of a blue readout)
							if((SciencePacket->Block3.DataBlockId==1)||(SciencePacket->Block3.DataBlockId==5)) 
							{
								//if we have already buffered some data, compress them now
								if (DmcPacketCounter > 0)
								{
									//compress and send the last buffer
									CompressLastPhotoBuffer();
									StoreNow = K_NO;
								}
								CurrentCompressionMode = SciencePacket->CompressionMode;  //store the current compression mode
							}
						}
					}

					//if we have not started buffering the data or if we are at the beginning of a new buffer
					if (DmcPacketCounter == 0)
					{
						if (CurrentCompressionMode == K_CM_PHOTO_BUFFER_TRANSMISSION) /* Case of Buffer Transmission Mode */
						{
							if (BufferTransmissionModeEnabled == K_NO)  //if the SPU is not currently transmitting the data from a previous buffer, we can start a new one
							{
								/* Store the Relevant Information of DEC/MEC Header from 1st Frame */
								CurrentObsId = SciencePacket->ObsId;		/* Get Initial Value of OBS ID */
								CswFifoMsg.Mode.DataBlockId = SciencePacket->Block3.DataBlockId;
								StoreNow = K_YES;						//start buffering the data
							}
						}
						else /* Case of Other Compression Modes */
						{
							/* Synchronize the beginning of a buffer with the first packet of a readout */
							if((SciencePacket->Block3.DataBlockId==1)||(SciencePacket->Block3.DataBlockId==5)) 
							{
								/* Store the Relevant Information of DEC/MEC Header from 1st Frame */
								CurrentObsId = SciencePacket->ObsId; /* Get Initial Value of OBS ID */
								CswFifoMsg.Mode.DataBlockId=SciencePacket->Block3.DataBlockId;
								StoreNow = K_YES;
							}
						}
					}

					//if we are now buffering the data
					if (StoreNow==K_YES)
					{
						//check if this is the first packet of a readout (Blue readouts have 4 packets)
						if((SciencePacket->Block3.DataBlockId==1)||(SciencePacket->Block3.DataBlockId==5))
						{
							//First, check if a new sequence just started on DMC.  In this case, we have to start a new buffer (if there was some readouts in the previous buffer).
							//inside sequences, the bit 0 of label is set to 1.
							//Note: this rule does not apply if we are in buffer transmission mode
							// in addition to that, there is now a sync delay feature that delays the start of a sequence by n frames.
						    if (((CurrentDmcLabel & 0x01) == 0) && ((SciencePacket->Label & 0x01) == 0x01))
						    {
						    	StartOfSequence = 1;
						    	Delay = DetectorConstantsPhoto.SyncDelay;
						    }

						    if (StartOfSequence == 1)
						    {
						    	if (Delay == 0)
						    	{
						    		if ((DmcPacketCounter != 0) && (CurrentCompressionMode != K_CM_PHOTO_BUFFER_TRANSMISSION))
						    		{
										CompressLastPhotoBuffer();
						    		}
	
						    		StartOfSequence = 0;
						    	}
							    else
							    {
							    	Delay--;
							    }
						    }

			    
						    CurrentDmcLabel = SciencePacket->Label;

							//store dmc header and data in circular buffers
							//note, the dmc header is only stored once per readout since it is the same in each packet (for each sub-array)
							if (WriteInCircularBuffer(K_CBUF_HEADER, &(SciencePacket->ObsId), K_PHOTO_HEADER_SIZE) == 0)
							{
								DiscardCurrentPhotoBuffers(SciencePacket->SpuId);
								PushDmcHeaderError(K_DMC_UNABLE_TO_PROCESS_DATA_FAST_ENOUGH);
								continue; //wait for next buffer
							}
							DmcHeaderStored++;
						}

						if (WriteInCircularBuffer(K_CBUF_SCIENCE_BLUE_PHOT1 + SciencePacket->Block3.DataBlockId - 1, SciencePacket->Data, K_NB_PIXELS_IN_PHOTO_ARRAY) == 0)
						{
							DiscardCurrentPhotoBuffers(SciencePacket->SpuId); 
							PushDmcHeaderError(K_DMC_UNABLE_TO_PROCESS_DATA_FAST_ENOUGH);
							continue; //wait for next buffer
						}
					
						DmcPacketCounter++;		//Increment the Packet Counter
						DmcReadoutsStored[SciencePacket->Block3.DataBlockId]++;	//Increment the readout counter
						
						/* Start Compression after Receiving 480 Frames */	
						if (DmcPacketCounter == K_NB_PHOTO_FRAMES_IN_3SEC_BUF)
						{
							CompressLastPhotoBuffer();

							//stop buffering.  It will restart automatically when everything is in condition again (this might well be at next frame) 
							StoreNow = K_NO;	
						}
					}
				};break;
				default: /*Invalid Data Type*/
				{
					PushDmcHeaderError(K_DMC_INVALID_DATA_TYPE);
					continue;  //ignore current packet and listen to next packet
				}break;
			}
		}
		else 
		{
			if (DmcPacketCounter != 0) // if we have just stopped the compression software, close or discar the latest buffers
			{
				//if we were in buffer transmission mode, stop the buffering (and the transmission will also stop (see fill_out.c))
				if ((CurrentCompressionMode == K_CM_PHOTO_BUFFER_TRANSMISSION) || (CurrentCompressionMode == K_CM_SPECTRO_BUFFER_TRANSMISSION))
				{
					//clear all buffers
					CircularBufferRemoveLastWrittenValues(K_CBUF_HEADER, DmcHeaderStored * K_PHOTO_HEADER_SIZE);
					CircularBufferRemoveLastWrittenValues(K_CBUF_SCIENCE_BLUE_PHOT1, DmcReadoutsStored[1] * K_NB_PIXELS_IN_PHOTO_ARRAY);
					CircularBufferRemoveLastWrittenValues(K_CBUF_SCIENCE_BLUE_PHOT2, DmcReadoutsStored[2] * K_NB_PIXELS_IN_PHOTO_ARRAY);
					CircularBufferRemoveLastWrittenValues(K_CBUF_SCIENCE_BLUE_PHOT3, DmcReadoutsStored[3] * K_NB_PIXELS_IN_PHOTO_ARRAY);
					CircularBufferRemoveLastWrittenValues(K_CBUF_SCIENCE_BLUE_PHOT4, DmcReadoutsStored[4] * K_NB_PIXELS_IN_PHOTO_ARRAY);
					CircularBufferRemoveLastWrittenValues(K_CBUF_SCIENCE_RED_PHOT, DmcReadoutsStored[5] * K_NB_PIXELS_IN_PHOTO_ARRAY);
					CircularBufferRemoveLastWrittenValues(K_CBUF_SCIENCE_SPECTRO, DmcReadoutsStored[0] * K_SPECTRO_FRAME_SIZE_IN_WORDS);
					DmcPacketCounter = 0;
					DmcHeaderStored = 0;
					for (i=0; i < 6; i++)
						DmcReadoutsStored[i] = 0;
				}
				else if (CurrentCompressionMode >= K_CM_SPECTRO_DEFAULT) //if we were in spectro
				{
					CompressLastSpectroBuffer();
				}
				else
				{
					CompressLastPhotoBuffer();
				}
			}

			CurrentCompressionMode = K_CM_UNDEFINED;	//while compression software is not running, reset the current compression mode
			StoreNow = K_NO;
			Delay = 0;
			StartOfSequence = 0;
		}
	}	
}

/*
This function triggers the compression of all the recently recorded spectro readouts
*/
void CompressLastSpectroBuffer()
{
	int i=0;

	//store the information that will be used by the compression task
	CswFifoMsg.Mode.ObsId = CurrentObsId;	// Get Initial Value of OBS ID
	CswFifoMsg.Mode.CompressionMode = CurrentCompressionMode;
	CswFifoMsg.Mode.ReadoutsPerRamps = SciencePacket->Block2.ReadoutsPerRamps;
	CswFifoMsg.Mode.PacketIndex = CompressedEntityCounter++;
	CswFifoMsg.Mode.DetectorType = K_SPECTRO;

	// we should make sure that we have a number of readouts that is a multiple of the number of readouts per sub-ramps
	CswFifoMsg.NbFrames = DmcReadoutsStored[0] - (DmcReadoutsStored[0] % DetectorConstantsSpectro.ReadoutsPerSubRamps);
	CircularBufferRemoveLastWrittenValues(K_CBUF_HEADER, (DmcHeaderStored - CswFifoMsg.NbFrames) * K_SPECTRO_HEADER_SIZE);
	CircularBufferRemoveLastWrittenValues(K_CBUF_SCIENCE_SPECTRO, (DmcReadoutsStored[0] - CswFifoMsg.NbFrames) * K_SPECTRO_FRAME_SIZE_IN_WORDS);

	//if we are in buffer transmission mode and if this is the last buffer of the 10 sec buffer, flag the buffer transmission mode as 'enabled'; this will prevent buffering readouts until the current buffer has been transmitted
	if ((CurrentCompressionMode == K_CM_SPECTRO_BUFFER_TRANSMISSION) && (BufferTransmissionModeBufferId == 4))
	{
		BufferTransmissionModeEnabled = K_YES;
	}
	
	DmcPacketCounter = 0;
	DmcHeaderStored = 0;
	for (i=0; i < 6; i++)
	{
		DmcReadoutsStored[i] = 0;
	}

	/* Start Compression of this Buffer*/
	if (CswFifoMsg.NbFrames > 0)
	{	
		KS_FIFOPut(FIFO_CSW, &CswFifoMsg); 
	}
}

/*
This function triggers the compression of all the recently recorded photo readouts
*/
void CompressLastPhotoBuffer()
{
	int i=0;

	//store the information that will be used by the compression task
	CswFifoMsg.Mode.CompressionMode = CurrentCompressionMode;
	CswFifoMsg.Mode.BolcDataType = (SciencePacket->Block1.BolcStatus >>7)&0x7; /* BOLC Status (keep only the 3 msb of 'data type') */
	CswFifoMsg.Mode.ObsId = CurrentObsId;
	CswFifoMsg.Mode.PacketIndex = CompressedEntityCounter++;
	CswFifoMsg.Mode.DetectorType = K_PHOTO;
	if (SciencePacket->SpuId == 2)	//DmcPacketCounter must be divided by 4 in case of blue photometry (DmcPacketCounter was counting the packets and nbFrames contains the number of readouts/frames)
	{
		//we should also make sure that we have the same number of readouts for each sub-array (this might not be the case when we receive a STOP CSW command or when we change the compression mode)
		int nbCompleteReadoutsStored = DmcReadoutsStored[4];
		CircularBufferRemoveLastWrittenValues(K_CBUF_HEADER, (DmcHeaderStored - nbCompleteReadoutsStored) * K_PHOTO_HEADER_SIZE);
		CircularBufferRemoveLastWrittenValues(K_CBUF_SCIENCE_BLUE_PHOT1, (DmcReadoutsStored[1] - nbCompleteReadoutsStored) * K_NB_PIXELS_IN_PHOTO_ARRAY);
		CircularBufferRemoveLastWrittenValues(K_CBUF_SCIENCE_BLUE_PHOT2, (DmcReadoutsStored[2] - nbCompleteReadoutsStored) * K_NB_PIXELS_IN_PHOTO_ARRAY);
		CircularBufferRemoveLastWrittenValues(K_CBUF_SCIENCE_BLUE_PHOT3, (DmcReadoutsStored[3] - nbCompleteReadoutsStored) * K_NB_PIXELS_IN_PHOTO_ARRAY);

		CswFifoMsg.NbFrames = nbCompleteReadoutsStored;
	}
	else
	{
		int nbCompleteReadoutsStored = DmcReadoutsStored[5];
		CircularBufferRemoveLastWrittenValues(K_CBUF_HEADER, (DmcHeaderStored - nbCompleteReadoutsStored) * K_PHOTO_HEADER_SIZE);

		CswFifoMsg.NbFrames = nbCompleteReadoutsStored;
	}

	//if we are in buffer transmission mode and if this is the last buffer of the 15 sec buffer, flag the buffer transmission mode as 'enabled'; this will prevent buffering readouts until the current buffer has been transmitted
	if ((CurrentCompressionMode == K_CM_PHOTO_BUFFER_TRANSMISSION) && (BufferTransmissionModeBufferId == 4))
	{
		BufferTransmissionModeEnabled = K_YES;
	}
	
	DmcPacketCounter = 0;
	DmcHeaderStored = 0;
	for (i=0; i < 6; i++)
	{
		DmcReadoutsStored[i] = 0;
	}

	/* Start Compression of this Buffer*/
	if (CswFifoMsg.NbFrames > 0)
	{
		KS_FIFOPut(FIFO_CSW, &CswFifoMsg); 
	}
}


/*
This function deletes all the recently recorded photo readouts
*/
void DiscardCurrentPhotoBuffers(u4 spuId)
{
	int i=0;
	
	CircularBufferRemoveLastWrittenValues(K_CBUF_HEADER, DmcHeaderStored * K_PHOTO_HEADER_SIZE);
	
	if (spuId == 2)
	{
		CircularBufferRemoveLastWrittenValues(K_CBUF_SCIENCE_BLUE_PHOT1, DmcReadoutsStored[1] * K_NB_PIXELS_IN_PHOTO_ARRAY);
		CircularBufferRemoveLastWrittenValues(K_CBUF_SCIENCE_BLUE_PHOT2, DmcReadoutsStored[2] * K_NB_PIXELS_IN_PHOTO_ARRAY);
		CircularBufferRemoveLastWrittenValues(K_CBUF_SCIENCE_BLUE_PHOT3, DmcReadoutsStored[3] * K_NB_PIXELS_IN_PHOTO_ARRAY);
		CircularBufferRemoveLastWrittenValues(K_CBUF_SCIENCE_BLUE_PHOT4, DmcReadoutsStored[4] * K_NB_PIXELS_IN_PHOTO_ARRAY);
	}
	else
	{
		CircularBufferRemoveLastWrittenValues(K_CBUF_SCIENCE_RED_PHOT, DmcReadoutsStored[5] * K_NB_PIXELS_IN_PHOTO_ARRAY);
	}
	
	DmcPacketCounter = 0;
	DmcHeaderStored = 0;
	
	for (i=0; i < 6; i++)
	{
		DmcReadoutsStored[i] = 0;
	}

	StoreNow = K_NO;
	BufferTransmissionModeBufferId = 0;		//reset the buffer id (used to store the buffer in PRAM when in buffer transmission mode)
	BufferTransmissionModeEnabled = K_NO;	//stop the buffer transmission mode
}

/*
This function deletes all the recently recorded spectro readouts
*/
void DiscardCurrentSpectroBuffers()
{
	int i=0;

	CircularBufferRemoveLastWrittenValues(K_CBUF_HEADER, DmcHeaderStored * K_SPECTRO_HEADER_SIZE);
	CircularBufferRemoveLastWrittenValues(K_CBUF_SCIENCE_SPECTRO, DmcReadoutsStored[0] * K_SPECTRO_FRAME_SIZE_IN_WORDS);
	DmcPacketCounter = 0;
	DmcHeaderStored = 0;
	
	for (i=0; i < 6; i++)
	{
		DmcReadoutsStored[i] = 0;
	}
	
	StoreNow = K_NO;
	BufferTransmissionModeBufferId = 0;		//reset the buffer id (used to store the buffer in PRAM when in buffer transmission mode)
	BufferTransmissionModeEnabled = K_NO;	//stop the buffer transmission mode
}

