/***************************************************************************************
* File name: Write.c
* Purpose:  Handles the Write Commands
*
* Description:	This Program Acknowledges the Reception of the Write Command
*				and Starts the Appropriate Program in Case of Command
*				Consistency.
****************************************************************************************/

#ifdef SIMULATOR
#include "virtuosoSim.h"
#else // SIMULATOR
#include "node1.h"
#include "k_struct.h"
#endif

#include "spulib.h"
#include "spuAsw.h"

/*=================================================================================================*/
/*Functions declared in this file
===================================================================================================*/

void Write(u4 buf[]);
void InitDetectorSelectionTables();
void RebuildDetectorSelectionTable(u4 nbPixels,  u4 *src, u4 *dest);
void InitTables(void);
void _MemCopyDmToDm(u4 *src, u4 *dest, u4 sizeInWords);

/*=================================================================================================*/
/*Global variables declared in this file
===================================================================================================*/

u4					WriteTempBuffer[512];	//intermediate buffer used for write commands

DXS_TABLE			DxsSpectro;
DXS_TABLE			DxsPhotoBlue;
DXS_TABLE			DxsPhotoRed;

WRT_DET_CST_SPECTRO	DetectorConstantsSpectro;
WRT_DET_CST_PHOTO	DetectorConstantsPhoto;
WRT_SIM_DATA		SimulationDataParameters;
u4					PreProcessingParametersSummary;
u4					BitMask1[32];		//array to ease some bit operations BitMask1[0] = 0x80000000, BitMask1[31] = 0x00000001
u4					BitMask2[32];		//array to ease some bit operations BitMask2[0] = 0x7FFFFFFF, BitMask2[31] = 0xFFFFFFFE
u4					WordMaskRight[32];	//array to ease some bit operations WordMaskRight[0] = 0x00000000, WordMaskRight[1] = 0x80000000, WordMaskRight[31] = 0xFFFFFFFE

u4 DetectorSelectionId = 1;				//the identifier of the current detector selection table (this variable is updated through a write command from DPU)
float SumOfX[512], LsqDenominator[512];	//these arrays are used in the ramp fitting algo
int RoundingHelper[16];					//this array is used in the resolution reduction algo (when using rounding method)
u4 PixelsPositionInSpectroArray[450];	//this array is used in detector selection (spectro).  It contains the position of pixels in the array (it avoids the 26th module that is not there)

/*=================================================================================================*/
/*External function and variables
===================================================================================================*/

extern u4 CbufFreeSpaceSummary;

/*================================================================================================*/


/*
This function handles the write commands from DPU
*/
void Write(u4 buf[])
{
	struct FIFO_TRANSMIT_MSG fifoTransmitMsg;	// message to send information to the data_tx task that will send the acknowledge to the DPU

	u4 parameterID, i, nbDetectors;
	u4 tempBuffer[0x18];
	u4 *address = 0x00000000;
	u4 lengthReceived, checksumReceived, computedChecksum;
	BOOL handleTheCommand = TRUE;

	/* Read Parameter ID */
   	parameterID = (buf[1]>>16)&0xFFFF;
	
	switch (parameterID)
	/* map the parameter ID to an address in RAM */					
	{
		case 0x81 : /* Case of the Write of the Table of Selected Detectors 1 to the SPU RAM */							
			address = DxsPhotoBlue.Table;
			break;
		case 0x82 : /* Case of the Write of the Table of Selected Detectors 2 to the SPU RAM */							
			address = &(DxsPhotoBlue.Table[K_NB_PIXELS_IN_PHOTO_BLOCK]);
			break;
		case 0x83 : /* Case of the Write of the Table of Selected Detectors 3 to the SPU RAM */							
			address = &(DxsPhotoBlue.Table[2*K_NB_PIXELS_IN_PHOTO_BLOCK]);
			break;
		case 0x84 : /* Case of the Write of the Table of Selected Detectors 4 to the SPU RAM */							
			address = &(DxsPhotoBlue.Table[3*K_NB_PIXELS_IN_PHOTO_BLOCK]);
			break;
		case 0x85 : /* Case of the Write of the Table of Selected Detectors 5 to the SPU RAM */							
            address = DxsPhotoRed.Table;
			break;
		case 0x86 : /* Case of the Write of the Table of Selected Detectors 6 to the SPU RAM */							
            address = DxsSpectro.Table;														 
			break;
		case 0x87 : /* Case of the Write of the Table of Selected Detectors 7 to the SPU RAM */							
            address = WriteTempBuffer;	//this DXS table was not used in the code so, ignore the write commands
			break;
		case 0x42 : /* Case of the Write of the Detectors Constants in Spectroscopy */
            address = (u4*)&DetectorConstantsSpectro;
			break;
		case 0x24 : /* Case of the Load of the Detectors Constants in Photometry */
            address = (u4*)&DetectorConstantsPhoto;
			break;
		case 0x18 : /* Case of the Load of the Needed Parameters for Data Simulation */
			address = (u4*)&SimulationDataParameters;
			break;
		default :  /* Invalid Parameter ID */
			{
				fifoTransmitMsg.Header = K_NACK_WRITE | K_INVALID_PARAMETER_ID; /* Invalid Parameter ID */
	  			fifoTransmitMsg.Block1.NackParameter = parameterID;
			    handleTheCommand = FALSE;
			}break;
	}

	if (handleTheCommand == TRUE)
	{
		/* Read the Write Parameters (Data Length and Checksum) */
	    lengthReceived = buf[1]&0xFFFF;
		checksumReceived = (buf[lengthReceived+2]>>16)&0xFFFF;
		
		/* The Length Granularity is 4-byte Words. Check the Length First 
			to Avoid Buffer Overflow*/
		if (lengthReceived != K_WRITE_CMD_LENGTH)
		{ 
			/* Invalid Length */
			fifoTransmitMsg.Header = K_NACK_WRITE | K_INVALID_LENGTH;
			fifoTransmitMsg.Block1.NackParameter = lengthReceived;
		}
		else
		{
			for (i=0;i<lengthReceived;i++)
			{
				tempBuffer[i] = buf[i+2];
			}

			computedChecksum=Crc(tempBuffer,lengthReceived*2,32);
			
			if (checksumReceived != computedChecksum)  /* Verify the Checksum */
			{ 
				/* Invalid Checksum */
				fifoTransmitMsg.Header = K_NACK_WRITE | K_INVALID_CHECKSUM;
				fifoTransmitMsg.Block1.NackParameter = computedChecksum;
			}
			else /* Positive Acknowledgment */		
			{
				/* In Case of DXS Table, Store the DXS ID and Setup Appropriate Table */
				if(parameterID>0x80) 
				{	
					DetectorSelectionId = tempBuffer[0];
					if(parameterID > 0x85)
					{
						if (parameterID == 0x86)
						{
							if(tempBuffer[2] > K_NB_USEFUL_PIXELS_IN_SPECTRO_ARRAY)
							{
								tempBuffer[2] = K_NB_USEFUL_PIXELS_IN_SPECTRO_ARRAY;
							}
							nbDetectors = K_NB_PIXELS_IN_SPECTRO_ARRAY;

							DxsSpectro.NbSelectedPixels = tempBuffer[2];
						}
						else
						{
							nbDetectors = 0; //table 0x87 is not used
						}
					}
					else
					{
						if(tempBuffer[2]>K_NB_PIXELS_IN_PHOTO_BLOCK)
						{
							tempBuffer[2]=K_NB_PIXELS_IN_PHOTO_BLOCK;
						}
						nbDetectors = K_NB_PIXELS_IN_PHOTO_BLOCK;

						if(parameterID==0x85)
						{
							DxsPhotoRed.NbSelectedPixels = tempBuffer[2];
						}
						else
						{
							DxsPhotoBlue.NbSelectedPixelsInEachPhotoBlock[parameterID - 0x81] = tempBuffer[2];
							DxsPhotoBlue.NbSelectedPixels = DxsPhotoBlue.NbSelectedPixelsInEachPhotoBlock[0] + DxsPhotoBlue.NbSelectedPixelsInEachPhotoBlock[1] + DxsPhotoBlue.NbSelectedPixelsInEachPhotoBlock[2] + DxsPhotoBlue.NbSelectedPixelsInEachPhotoBlock[3];
						}
					}

					RebuildDetectorSelectionTable(nbDetectors, tempBuffer+3, address);

					/* Positive Acknowledgment */
					fifoTransmitMsg.Header = K_PACK_WRITE;
				}
				else
				{ 				
					/* Copy Data to target */
					MemCopyDmToDm(tempBuffer, address,lengthReceived);

					/*Store the Sub-ramp Length*/
					if(parameterID==0x42) 
					{
						/*Check that NbSamplesPerSubRamps is Valid (2,4,8,16,32,..,512) */
						if(DetectorConstantsSpectro.ReadoutsPerSubRamps <= 1)
						{
							/*Negative Acknowledgment Invalid Parameter*/
							fifoTransmitMsg.Header = K_NACK_WRITE | K_INVALID_SID_PARAMETER;
							fifoTransmitMsg.Block1.NackParameter = DetectorConstantsSpectro.ReadoutsPerSubRamps;
						}
						/*Check that PACS Codec Order is valid (3 or 4) */
						else if ((DetectorConstantsSpectro.PacsCodecOrder < 3) || (DetectorConstantsSpectro.PacsCodecOrder > 4))
						{
							/*Negative Acknowledgment Invalid Parameter*/
							fifoTransmitMsg.Header = K_NACK_WRITE | K_INVALID_SID_PARAMETER;
							fifoTransmitMsg.Block1.NackParameter = DetectorConstantsSpectro.PacsCodecOrder;
						}
						else
						{
							u4 nbBitsAtOne = 0;
						
							//check that the ReadoutsPerSubRamps is a power of 2
							for (i=1;i<11;i++)
							{
								nbBitsAtOne += ((DetectorConstantsSpectro.ReadoutsPerSubRamps >> i) & 0x1);
							}

							nbBitsAtOne = 1; // deactivate power of 2 check for v13.96

							if((nbBitsAtOne==1)&&(DetectorConstantsSpectro.ReadoutsPerSubRamps < 513))
							{
								/*Store the Number of Sub-Ramps and shift Nb and Send Pack*/
								for (i=1;i<11;i++)
								{
									if(((DetectorConstantsSpectro.ReadoutsPerSubRamps >> i) & 0x1) == 1)
									{
										DetectorConstantsSpectro.Log2OfReadoutsPerSubRamps = i;
									}
								}
					
								PreProcessingParametersSummary = ((((DetectorConstantsSpectro.RoundingBits<<4) | DetectorConstantsSpectro.SortingAlgorithm)<<16) & 0xFFFF0000) | (DetectorConstantsSpectro.LosslessCompressionAlgorithm & 0xFFFF);	
								fifoTransmitMsg.Header = K_PACK_WRITE;
							}
							else
							{
								/*Negative Acknowledgment Invalid Parameter*/
								fifoTransmitMsg.Header = K_NACK_WRITE | K_INVALID_SID_PARAMETER;
								fifoTransmitMsg.Block1.NackParameter = DetectorConstantsSpectro.ReadoutsPerSubRamps;
							}
						}

					}
					else
					{
						fifoTransmitMsg.Header = K_PACK_WRITE;
					}
				}	
			}
		}
	}

	KS_FIFOPut(FIFO_TRANSMIT, &fifoTransmitMsg); /* Send the Response Packet to the DPU */

}

/*
This function fills the detector selection tables with default values (all pixels selected)
*/
void InitDetectorSelectionTables()
{
	u4 i;
	
	/* Initialize Detector selection table in Photometry */
	
	for (i=0;i<4;i++)
	{
		DxsPhotoBlue.NbSelectedPixelsInEachPhotoBlock[i]= K_NB_PIXELS_IN_PHOTO_BLOCK;
	}
	
	DxsPhotoBlue.Size = K_NB_PIXELS_IN_BLUE_BOLOMETERS;
	DxsPhotoBlue.NbSelectedPixels = K_NB_PIXELS_IN_BLUE_BOLOMETERS;
	for (i=0;i<K_NB_PIXELS_IN_BLUE_BOLOMETERS;i++)
	{
		DxsPhotoBlue.Table[i]= 1;
	}
	
	DxsPhotoRed.Size = K_NB_PIXELS_IN_PHOTO_BLOCK;
	DxsPhotoRed.NbSelectedPixels = K_NB_PIXELS_IN_PHOTO_BLOCK;
	
	for (i=0;i<K_NB_PIXELS_IN_PHOTO_BLOCK;i++)
	{
		DxsPhotoRed.Table[i]= 1;
	}
	
	/* Initialize Detector selection table in Spectroscopy */
	DxsSpectro.NbSelectedPixels = K_NB_USEFUL_PIXELS_IN_SPECTRO_ARRAY;
	DxsSpectro.Size = K_NB_PIXELS_IN_SPECTRO_ARRAY;
	
	for(i=0;i<K_NB_PIXELS_IN_SPECTRO_ARRAY;i++)
	{
		DxsSpectro.Table[i] = 1;
	}
	
	for(i=1;i<=18;i++)
	{
		DxsSpectro.Table[(i*26)-1] = 0;
	}
}


/*
This function rebuilds a detector selection table after a write command.
*/
void RebuildDetectorSelectionTable(u4 nbPixels, u4 *src, u4 *dest)
{
	u4 pixel, granul=0;
	const int K_TAB_GRANULARITY = 32;
	const u4 K_MSB = 0x80000000;
	
	for(pixel = 0; pixel<nbPixels; pixel++, granul<<=1) 
	{
		/* Is the actual granul valid? */
		if (pixel % K_TAB_GRANULARITY == 0)
		{
			granul = src[pixel/K_TAB_GRANULARITY];
		}

		/* Is the pixel valid? */
		if ((granul & K_MSB) == K_MSB)
		{
			dest[pixel] = 1;
		}
		else
		{
			dest[pixel] = 0;
		}
	}
}

/*
This function initializes many tables:
- it writes default values in the detector constants and simulation parameters
- it builds some helper tables: RoundingHelper, BitMask, ...
*/
void InitTables(void)
{
	u4 i, j, k;
	u4 pixel;
	float sumOfXbyX[512], sumOfXSquare[512];

	SimulationDataParameters.SpuId = 3;
	SimulationDataParameters.CompressionMode = K_CM_SPECTRO_DEFAULT;
	SimulationDataParameters.ReadoutsPerRamps = 0x32;
	SimulationDataParameters.ProgrammableHkAddress = (u4)&CbufFreeSpaceSummary;	
	SimulationDataParameters.ComputeChecksumOnScienceData = 1;
	SimulationDataParameters.DmcHeaderCompressionAlgorithm = 0; // 0 = new (rzip2), 1 = old (rzip32)

	/*=============================================================================================*/
	/* Detectors Constants Tables Initialization */
	DetectorConstantsSpectro.PreProcessingFlag = 0;
	DetectorConstantsSpectro.PreProcessingParameter = 0;
	DetectorConstantsSpectro.ThresholdForSpectro = 1000;
	DetectorConstantsSpectro.ThresholdForChopper = 100;
	DetectorConstantsSpectro.ThresholdForGrating = 100;
	DetectorConstantsSpectro.UseGlitchDetection = K_NO;
	DetectorConstantsSpectro.RampFittingAlgorithm = K_MEAN_ALGO; //K_LEAST_SQUARE_ALGO;
	DetectorConstantsSpectro.SortingAlgorithm = 5;
	DetectorConstantsSpectro.LosslessCompressionAlgorithm = 4;
	DetectorConstantsSpectro.PacsCodecOrder = 3;
	DetectorConstantsSpectro.ReadoutsPerSubRamps = 8;
	DetectorConstantsSpectro.Log2OfReadoutsPerSubRamps = 3;
	DetectorConstantsSpectro.RoundingBits = 0;
	DetectorConstantsSpectro.RoundingStrategy = 0;
	DetectorConstantsSpectro.FixRawChannels = 0;
	DetectorConstantsSpectro.DiscardPre = 0;
	DetectorConstantsSpectro.DiscardPost = 0;
#ifdef SIMULATOR
	DetectorConstantsSpectro.ReadoutsPerSubRamps = 8;
	DetectorConstantsSpectro.Log2OfReadoutsPerSubRamps = 3;
	DetectorConstantsSpectro.SortingAlgorithm = 5;
	DetectorConstantsSpectro.LosslessCompressionAlgorithm = 4;
	DetectorConstantsSpectro.PacsCodecOrder = 3;
	RawChannelSelectionSpectro.NbSelectedRawChannels = 2;
	DetectorConstantsSpectro.RoundingBits = 3;
	DetectorConstantsSpectro.RoundingStrategy = 1;
	DetectorConstantsSpectro.FixRawChannels = 0;
#endif


	DetectorConstantsPhoto.PreProcessingFlag = 0;
	DetectorConstantsPhoto.PreProcessingParameter = 0;
	DetectorConstantsPhoto.ThresholdForPhoto = 1000;
	DetectorConstantsPhoto.ThresholdForChopper = 100;
	DetectorConstantsPhoto.UseGlitchDetection = K_NO;
	DetectorConstantsPhoto.NbSamplesToAverage = 0; // 0 means to decide by cmm and use int averaging
	DetectorConstantsPhoto.Scheme = 1; // 0=velveteen, 1=velvet
	DetectorConstantsPhoto.LosslessCompressionAlgorithm = 0;
	DetectorConstantsPhoto.PacsCodecOrder = 3;
	DetectorConstantsPhoto.RoundingBits = 0;
	DetectorConstantsPhoto.RoundingStrategy = 0;
	DetectorConstantsPhoto.DiscardPre = 0;
	DetectorConstantsPhoto.DiscardPost = 0;
	DetectorConstantsPhoto.SyncDelay = 0;
#ifdef SIMULATOR
	DetectorConstantsPhoto.Scheme = 1; // 0=velveteen, 1=velvet
	DetectorConstantsPhoto.LosslessCompressionAlgorithm = 0;
	DetectorConstantsPhoto.NbSamplesToAverage = 0; 
	RawChannelSelectionPhoto.NbSelectedRawChannels = 2;
	DetectorConstantsPhoto.RoundingBits = 3;
	DetectorConstantsPhoto.RoundingStrategy = 1;
#endif

	/* Initialization of Data Reduction Parameters to Default */
	if (DetectorConstantsPhoto.RoundingStrategy == K_ROUNDING_STYLE_138)
		PreProcessingParametersSummary = (((DetectorConstantsSpectro.SortingAlgorithm)<<16)&0xFFFF0000)|(DetectorConstantsSpectro.LosslessCompressionAlgorithm & 0xFFFF);	
	else
		PreProcessingParametersSummary = ((((DetectorConstantsSpectro.RoundingBits<<4) | DetectorConstantsSpectro.SortingAlgorithm)<<16)&0xFFFF0000)|(DetectorConstantsSpectro.LosslessCompressionAlgorithm & 0xFFFF);	
	/*=============================================================================================*/
    /* Bit Mask Tables Preparation */

	BitMask1[31] = 0x00000001;

	for(i=0; i < 31; i++)
	{
		BitMask1[i] = 0x00000001 << (31-i);
	}
	for(i=0; i < 32; i++)
	{
		BitMask2[i] = ~BitMask1[i];
		WordMaskRight[i] = 0xfffffffe<<(31-i);	
	}

	/*=============================================================================================*/
    /* Rounding Lookup Table Preparation */

	RoundingHelper[0]=0;
	RoundingHelper[1]=1;
	for(i=2; i<16; i++)
	{
		RoundingHelper[i] = RoundingHelper[i-1]*2;
	}
	
	/*===========================================*/
    /* Least Square Tables Preparation */

	SumOfX[0] = sumOfXSquare[0]= sumOfXbyX[0]=0;
	LsqDenominator[0]= 1.0f; /* To Avoid Division By Zero */

	for(i=1;i<512;i++)
	{
		SumOfX[i] = i + SumOfX[i-1];
		sumOfXSquare[i] =(i*i) + sumOfXSquare[i-1];
        sumOfXbyX[i]= SumOfX[i] * SumOfX[i];
		LsqDenominator[i]= ((i+1) * sumOfXSquare[i] - sumOfXbyX[i]);
	}

	/*=============================================================================================*/
	/* Lookup Table for fastening DXS spectroscopy*/

	pixel=0;
	k=0;
	for(i=0;i<18;i++)
	{
		for(j=0;j<25;j++)
		{
			PixelsPositionInSpectroArray[k++] = pixel++;
		}
		pixel++;
	}
}


/*
We have replaced the MEM_CopyDMToDM function provided by CRISA because it was incredibly slow.
*/
void _MemCopyDmToDm(u4 *src, u4 *dest, u4 sizeInWords)
{
	u4 i=0;

	for(i=0; i<sizeInWords; i++) 
	{
		dest[i] = src[i];
	}
	return;
}

